import shutil
import click
import subprocess
import tempfile
import os
from convisoappsec.flowcli.context import pass_flow_context
from datetime import datetime
from convisoappsec.flowcli.requirements_verifier import RequirementsVerifier
from convisoappsec.flowcli import help_option
from convisoappsec.flowcli.common import asset_id_option


@click.command()
@asset_id_option(required=False)
@click.option(
    '-r',
    '--repository-dir',
    default=".",
    show_default=True,
    type=click.Path(
        exists=True,
        resolve_path=True,
    ),
    required=False,
    help="The source code repository directory.",
)
@click.option(
    "--send-to-flow/--no-send-to-flow",
    default=True,
    show_default=True,
    required=False,
    help="""Enable or disable the ability of send analysis result
    reports to flow.""",
    hidden=True
)
@click.option(
    "--custom-sca-tags",
    hidden=True,
    required=False,
    multiple=True,
    type=(str, str),
    help="""It should be passed as <repository_name> <image_tag>. It accepts multiple values"""
)
@click.option(
    "--scanner-timeout",
    hidden=True,
    required=False,
    default=7200,
    type=int,
    help="Set timeout for each scanner"
)
@click.option(
    "--parallel-workers",
    hidden=True,
    required=False,
    default=2,
    type=int,
    help="Set max parallel workers"
)
@click.option(
    "--deploy-id",
    default=None,
    required=False,
    hidden=True,
    envvar=("CONVISO_DEPLOY_ID", "FLOW_DEPLOY_ID")
)
@click.option(
    '--experimental',
    default=False,
    is_flag=True,
    hidden=True,
    help="Enable experimental features.",
)
@click.option(
    "--company-id",
    required=False,
    envvar=("CONVISO_COMPANY_ID", "FLOW_COMPANY_ID"),
    help="Company ID on Conviso Platform",
)
@click.option(
    '--asset-name',
    required=False,
    envvar=("CONVISO_ASSET_NAME", "FLOW_ASSET_NAME"),
    help="Provides a asset name.",
)
@click.option(
    '--vulnerability-auto-close',
    default=False,
    is_flag=True,
    hidden=True,
    help="Enable auto fixing vulnerabilities on cp.",
)
@click.option(
    '--from-ast',
    default=False,
    is_flag=True,
    hidden=True,
    help="Internal use only.",
)
@click.option(
    '--control-sync-status-id',
    required=False,
    hidden=True,
    help="Control sync status id.",
)
@help_option
@pass_flow_context
@click.pass_context
def generate(context, flow_context, asset_id, company_id, repository_dir, send_to_flow, custom_sca_tags,
             scanner_timeout, parallel_workers, deploy_id, experimental, asset_name, vulnerability_auto_close,
             from_ast, control_sync_status_id):
    # Prepare context if not coming from AST
    if not from_ast:
        try:
            prepared_context = RequirementsVerifier.prepare_context(context)
        except Exception as e:
            log_func(f"⚠️ Error preparing context: {e}. Exiting.")
            return

        # Copy parameters from locals or prepared_context
        params_to_copy = [
            'asset_id', 'company_id', 'repository_dir', 'send_to_flow',
            'deploy_id', 'custom_sca_tags', 'scanner_timeout', 'parallel_workers',
            'experimental', 'asset_name', 'vulnerability_auto_close'
        ]
        for param_name in params_to_copy:
            context.params[param_name] = (
                locals()[param_name] or prepared_context.params[param_name]
            )

    # Generate SBOM file
    log_func("💬 Generating SBOM file...")
    try:
        asset_name = context.params['asset_name']
        temp_dir = tempfile.gettempdir()
        timestamp = datetime.now().strftime('%Y%m%d_%H%M%S')
        sanitized_asset_name = (asset_name or "").replace(" ", "_").replace("(", "").replace(")", "")
        file_name = os.path.join(temp_dir, f"sbom_{sanitized_asset_name}_{timestamp}.json")

        exclude_patterns = [
            "**/.github/**",
            "**/node_modules/**",
            "**/target/**",
            "**/vendor/**",
            "**/build/**",
            "**/dist/**",

            "**/test/**",
            "**/tests/**",
            "**/__pycache__/**",

            "**/*.dll",
            "**/*.exe",
            "**/*.so"
        ]

        exclude_string = ",".join(exclude_patterns)
        command = f'cdxgen -o {file_name} --exclude "{exclude_string}"'

        catalogers = [
            '-github-actions',
            '-python-installed-package-cataloger',
            '-sbom-cataloger',
            '-file-content-cataloger',
            '-file-digest-cataloger',
            '-file-executable-cataloger',
            '-file-metadata-cataloger'
        ]

        try:
            if command_exists('cdxgen'):
                subprocess.run(command, shell=True, check=True, capture_output=True)
            else:
                raise FileNotFoundError("cdxgen not found")

        except (subprocess.CalledProcessError, FileNotFoundError) as e:
            print(f"[!] cdxgen failed ({e}), falling back to syft...")
            subprocess.run(
                "curl -sSfL https://raw.githubusercontent.com/anchore/syft/main/install.sh | sh -s -- -b conviso/",
                shell=True,
                stdout=subprocess.DEVNULL,
                stderr=subprocess.DEVNULL
            )
            command = [f"./conviso/syft scan {repository_dir} -o cyclonedx-json={file_name} "
                       f"--select-catalogers '{','.join(catalogers)}'"]

            subprocess.run(command, shell=True, check=True, capture_output=True)

        directory = 'conviso/'
        if os.path.isdir(directory):
            shutil.rmtree(directory)

        log_func("✅ SBOM file generated successfully!")
    except subprocess.CalledProcessError as error:
        log_func(f"⚠️ Error generating SBOM file: {error}.")
        return
    except Exception as e:
        log_func(f"⚠️ Unexpected error during SBOM generation: {e}")
        return

    # Ensure asset_id and company_id is available
    asset_id = asset_id or context.params.get('asset_id')
    company_id = company_id or context.params.get('company_id')

    if not asset_id:
        log_func(f"⚠️ Missing asset_id. Unable to send SBOM.")
        return

    # Send SBOM file to CSC (Conviso Platform)
    try:
        send_sbom_file_to_csc(company_id=company_id, asset_id=asset_id, file=file_name)
    except Exception as e:
        log_func(f"⚠️ Error sending SBOM file to Conviso: {e}")
        return


@pass_flow_context
def send_sbom_file_to_csc(flow_context, company_id, asset_id, file):
    try:
        conviso_api = flow_context.create_conviso_graphql_client()
        api_key = flow_context.key

        log_func("💬 Sending SBOM to the Conviso Platform...")
        conviso_api.sbom.send_sbom_file(company_id=company_id, asset_id=asset_id, file_path=file, api_key=api_key)
        log_func("✅ SBOM file sent successfully!")
    except Exception as e:
        log_func(f"⚠️ Failed to send SBOM file: {e}")


def command_exists(command):
    return shutil.which(command) is not None


def log_func(msg, new_line=True):
    click.echo(click.style(msg), nl=new_line, err=True)



import re
from convisoappsec.flowcli.common import CreateDeployException
from convisoappsec.logger import LOGGER
from convisoappsec.flowcli.companies.ls import Companies
from convisoappsec.flow.graphql_api.v1.models.asset import AssetInput
from convisoappsec.common.git_data_parser import GitDataParser
from .context import pass_flow_context


class RequirementsVerifier:

    @staticmethod
    @pass_flow_context
    def list_assets(flow_context, company_id, asset_name, scan_type):
        conviso_api = flow_context.create_conviso_graphql_client()

        asset_model = AssetInput(
            int(company_id),
            asset_name,
            scan_type
        )

        return conviso_api.assets.list_assets(asset_model)

    @staticmethod
    @pass_flow_context
    def create_asset(flow_context, company_id, asset_name, scan_type):
        conviso_api = flow_context.create_conviso_graphql_client()

        asset_model = AssetInput(
            int(company_id),
            asset_name,
            scan_type
        )

        return conviso_api.assets.create_asset(asset_model)

    @staticmethod
    def sarif_asset_assignment(context, asset):
        """ assignment asset when is a sarif import """
        context.params['asset_id'] = asset['id']
        context.params['experimental'] = True

        return context

    @staticmethod
    def find_or_create_asset(context, company_id, old_name, new_name):
        """ Method to find or create asset on conviso platform """
        try:
            existing_assets = RequirementsVerifier.list_assets(company_id, new_name, 'SAST')
            if not existing_assets:
                existing_assets = RequirementsVerifier.list_assets(company_id, old_name, 'SAST')
            for asset in existing_assets:
                if asset['name'] == old_name or asset['name'] == new_name:
                    LOGGER.info('✅ Asset found...')
                    context.params['asset_name'] = asset['name']
                    return [asset]
            LOGGER.info('💬 Asset not found; creating...')
            new_asset = RequirementsVerifier.create_asset(company_id, new_name, 'SAST')
            context.params['asset_name'] = new_name
            return [new_asset]
        except Exception as e:
            raise Exception("Error: {}".format(e))

    @staticmethod
    def create_asset_with_custom_name(context, company_id, asset_name):
        """ Create an asset with custom name pass with a custom name """
        if not asset_name or not asset_name.strip():  # Check for None or blank string
            raise ValueError("Asset name cannot be None or blank.")

        # we need to verify if already has an asset with the name provided.
        # because graphql will return an error if already has.
        existing_asset = RequirementsVerifier.list_assets(company_id, asset_name, 'SAST')

        if not existing_asset:
            LOGGER.info("💬 Asset not found; creating with name {}...".format(asset_name))
            asset = RequirementsVerifier.create_asset(company_id, asset_name, 'SAST')
        else:
            LOGGER.info('✅ Asset found...')
            asset = existing_asset[0]

        context.params['asset_name'] = asset_name

        return asset

    @staticmethod
    @pass_flow_context
    def prepare_context(flow_context, context, from_ast=False):
        """ Due to the new vulnerability management we need to do some checks before continuing the flow """

        if from_ast is True:
            context.params['from_ast'] = True

        companies = Companies()
        company_id = context.params['company_id']

        if company_id is not None:
            companies_filtered = [companies.ls(flow_context, company_id=company_id)]
        else:
            companies_filtered = companies.ls(flow_context)

        if len(companies_filtered) > 1:
            raise CreateDeployException(
                "❌ Deploy not created. You have access to multiple companies; please specify one using CONVISO_COMPANY_ID."
            )

        company = companies_filtered[0]
        company_id = company['id']

        if context.params['asset_name'] is not None:
            # if user use --asset-name param or envvar CONVISO_ASSET_NAME, FLOW_ASSET_NAME
            asset_name = context.params['asset_name']
            asset = RequirementsVerifier.create_asset_with_custom_name(context, company_id, asset_name)
        else:
            pattern = r"\([^)]*\)"  # eliminating what is in parentheses
            old_asset_name = GitDataParser(context.params['repository_dir']).parse_name()
            new_asset_name = re.sub(pattern, '', old_asset_name).strip()

            assets = RequirementsVerifier.find_or_create_asset(context, company_id, old_asset_name, new_asset_name)
            asset = assets[0]

        if 'input_file' in context.params:
            # sarif only uses assets, not requiring the creation of a project.
            RequirementsVerifier.sarif_asset_assignment(context, asset)

            return context

        context.params['asset_id'] = asset['id']
        context.params['experimental'] = True
        context.params['company_id'] = company_id

        return context



import click

from convisoappsec.flowcli import help_option

from .with_ import with_


@click.group()
@help_option
def create():
    pass


create.add_command(with_)

create.epilog = '''
  Run flow findings create COMMAND --help for more information on a command.
'''



import click

from convisoappsec.flowcli import help_option
from .create import create
from .import_sarif import import_sarif


@click.group()
@help_option
def findings():
    pass


findings.add_command(create)
findings.add_command(import_sarif)

findings.epilog = '''
  Run flow findings COMMAND --help for more information on a command.
'''



import click


def print_help(ctx, param, value):
    if not value or ctx.resilient_parsing:
        return
    click.echo(ctx.get_help())
    ctx.exit()


help_option = click.option(
    '-h', '--help',
    is_flag=True,
    callback=print_help,
    expose_value=False,
    is_eager=True,
    help="Show this message and exit."
)



from enum import IntEnum
import json

class Severity(IntEnum):
    # Sastbox [:undefined, :info, :low, :medium, :high, :critical]
    UNDEFINED = 0
    INFO = 1
    LOW = 2
    MEDIUM = 3
    HIGH = 4
    CRITICAL = 5
    
    @classmethod
    def has_value(cls, value):
        return value in cls._member_names_



class Decision:


    def __init__(self, report):
        self.report = open(report)
        self.json = json.load(self.report)
        self.issues = self.json.get('issues')
        self.sev = self.issues

    def block_from_severity(self, severity, threshold=1):
        if not severity:
            return False
        severity = Severity[severity]
        self.sev = [issue for issue in self.issues if Severity[issue.get('severity').upper()] >= severity]
        return len(self.sev) >= threshold

    def block_from_findings(self, threshold):
        if not threshold:
            return False
        return len(self.issues) >= threshold
    
    def filtered_issues(self):
        return self.sev







import click

from convisoappsec.flowcli import help_option
from .run import run


@click.group()
@help_option
def container():
    pass


container.add_command(run)

container.epilog = '''
  Run conviso container COMMAND --help for more information on a command.
'''



from .source_code_scanner import SourceCodeScanner # noqa
from .exceptions import SourceCodeScannerException # noqa
from .scc import SCC # noqa

__all_ = [
    'SourceCodeScanner',
    'SourceCodeScannerException',
    'SCC'
]



import jmespath
import json
import requests

from convisoappsec.flow.graphql_api.v1.models.asset import AssetInput
from convisoappsec.flow.graphql_api.v1.schemas import mutations, resolvers
from urllib.parse import urlparse
from convisoappsec.version import __version__


class AssetsAPI(object):
    """ To operations on Asset's resources in Conviso Platform. """

    def __init__(self, conviso_graphql_client):
        self.__conviso_graphql_client = conviso_graphql_client

    def create_asset(self, asset_input: AssetInput):
        graphql_variables = asset_input.to_graphql_dict()

        graphql_body_response = self.__conviso_graphql_client.execute(
            mutations.CREATE_ASSET,
            graphql_variables
        )

        expected_path = 'createAsset.asset'

        asset = jmespath.search(
            expected_path,
            graphql_body_response,
        )

        return asset

    def update_asset(self, company_id, asset_id, asset_name, technologies=None, repo_url=None):
        graphql_variables = {
            "id": asset_id,
            "companyId": company_id,
            "name": asset_name,
            "tecnologyList": technologies,
            "repoUrl": repo_url
        }

        graphql_body_response = self.__conviso_graphql_client.execute(
            mutations.UPDATE_ASSET,
            graphql_variables
        )

        expected_path = 'updateAsset.asset'

        asset = jmespath.search(
            expected_path,
            graphql_body_response,
        )

        return asset

    def get_by_company_id_or_name(self, company_id, asset_name, page, limit):
        graphql_variables = {
            "id": company_id,
            'name': asset_name,
            "page": page,
            "limit": limit
        }

        graphql_body_response = self.__conviso_graphql_client.execute(
            resolvers.GET_ASSETS,
            graphql_variables
        )

        expected_path = 'assets.collection'

        assets_by_company = jmespath.search(
            expected_path,
            graphql_body_response,
        )

        return assets_by_company

    def get_asset_url(self, company_id, asset_id):
        parsed_url = urlparse(self.__conviso_graphql_client.url)

        asset_path = '/scopes/{}/assets/{}'.format(
            company_id,
            asset_id
        )

        parsed_url = parsed_url._replace(path=asset_path)

        return parsed_url.geturl()

    def list_assets(self, params, page=1, limit=32):
        graphql_variables = {
            "id": params.company_id,
            "name": params.name,
            "page": page,
            "limit": limit
        }

        graphql_body_response = self.__conviso_graphql_client.execute(
            resolvers.GET_ASSETS,
            graphql_variables
        )

        expected_path = 'assets.collection'

        assets = jmespath.search(
            expected_path,
            graphql_body_response,
        )

        return assets

class CompaniesApi(object):
    """ To operations on Companies resources in Conviso Platform. """

    def __init__(self, conviso_graphql_client):
        self.__conviso_graphql_client = conviso_graphql_client

    def get_company_by_id(self, company_id):
        graphql_variables = {
            "company_id": company_id,
        }

        graphql_body_response = self.__conviso_graphql_client.execute(
            resolvers.GET_COMPANY,
            graphql_variables
        )

        expected_path = 'company'
        company = jmespath.search(
            expected_path,
            graphql_body_response,
        )
        return company

    def get_companies(self, page=1, limit=10):
        graphql_variables = {
            "page": page,
            "limit": limit
        }

        graphql_body_response = self.__conviso_graphql_client.execute(
            resolvers.GET_COMPANIES,
            graphql_variables
        )

        expected_path = 'companies.collection'

        companies = jmespath.search(
            expected_path,
            graphql_body_response,
        )

        return companies


class IssuesApi(object):
    """ To operations on Issue's resources in Conviso Platform. """

    def __init__(self, conviso_graphql_client):
        self.__conviso_graphql_client = conviso_graphql_client

    def get_issues_stats(self, asset_id, company_id, statuses, end_date=None):
        """ Return issue stats filter by asset and company """

        graphql_variables = {
            'asset_id': asset_id,
            'company_id': company_id,
            'statuses': statuses,
            'end_date': end_date
        }

        graphql_body_response = self.__conviso_graphql_client.execute(
            resolvers.GET_ISSUES_STATS,
            graphql_variables
        )

        expected_path = 'issuesStats.severities'

        issues_stats = jmespath.search(
            expected_path,
            graphql_body_response,
        )

        return issues_stats

    def send_issues_file(self, company_id, asset_id, file_path, api_key, vulnerability_type='SAST_FINDING',
                         deploy_id=None, commit_ref=None, control_sync_status_id=None):
        """Send issues file to Conviso platform"""

        url = self.__conviso_graphql_client.url

        if not isinstance(vulnerability_type, list):
            vulnerability_type = [vulnerability_type]

        vulnerability_type = [v.upper() for v in vulnerability_type]

        variables = {
            "companyId": company_id,
            "assetId": asset_id,
            "vulnerabilityTypes": vulnerability_type,
            "deployId": deploy_id,
            "commitRef": commit_ref,
            "controlSyncStatusId": control_sync_status_id,
            "file": None
        }

        operations = {
            "query": mutations.IMPORT_FINDINGS,
            "variables": variables
        }

        file_map = {"0": ["variables.file"]}

        headers = {
            'x-api-key': api_key,
            "User-Agent": f"AST:{__version__}"
        }

        try:
            with open(file_path, 'rb') as file:
                files = {
                    'operations': (None, json.dumps(operations), 'application/json'),
                    'map': (None, json.dumps(file_map), 'application/json'),
                    '0': (file_path, file, 'application/octet-stream')
                }

                response = requests.post(url, files=files, headers=headers)
                response.raise_for_status()
                json_response = response.json()

                self._handle_graphql_errors(json_response)
                return json_response.get('data')

        except requests.RequestException as e:
            raise Exception(f"GraphQL request failed: {e}")

    @staticmethod
    def _handle_graphql_errors(json_response):
        """Handle GraphQL errors"""
        errors = json_response.get('errors')
        if errors:
            error_messages = [error.get('message', 'Unknown error') for error in errors]
            raise Exception(f"GraphQL request failed with errors: {', '.join(error_messages)}")


class DeploysApi(object):
    """ Class for deploys resources """

    def __init__(self, conviso_graphql_client):
        self._conviso_graphql_client = conviso_graphql_client

    def get_deploys_by_asset(self, asset_id):
        """ Returns deploys based on the provided asset id """

        graphql_variables = {
            'asset_id': asset_id
        }

        graphql_body_response = self._conviso_graphql_client.execute(
            resolvers.GET_DEPLOYS_BY_ASSET,
            graphql_variables
        )

        expected_path = 'deploysByAsset.collection'

        deploys = jmespath.search(
            expected_path,
            graphql_body_response
        )

        return deploys

    def create_deploy(self, asset_id, previous_commit, current_commit, branch_name, api_key,
                      commit_history):
        """create a deployment"""
        url = self._conviso_graphql_client.url
        operations = {
            "query": mutations.CREATE_DEPLOY,
            "variables": {
                "assetId": asset_id,
                "previousCommit": previous_commit,
                "currentCommit": current_commit,
                "branchName": branch_name,
                "diffContent": None,
                "commitHistory": None
            }
        }
        file_map = {
            "0": ["variables.diffContent"],
            "1": ["variables.commitHistory"]
        }

        with open(commit_history, 'rb') as f:
            commit_history_data = f.read()

        files = {
            'operations': (None, json.dumps(operations), 'application/json'),
            'map': (None, json.dumps(file_map), 'application/json'),
            '0': (commit_history, commit_history_data, 'application/octet-stream'),
            '1': (commit_history, commit_history_data, 'application/octet-stream'),
        }

        headers = {
            'x-api-key': api_key,
            "User-Agent": f"AST:{__version__}"
        }

        response = requests.post(url, files=files, headers=headers)
        response.raise_for_status()
        json_response = response.json()
        self._handle_graphql_errors(json_response)
        return json_response.get('data')

    # TODO: extract
    @staticmethod
    def _handle_graphql_errors(json_response):
        """ Handle GraphQL errors """
        if 'errors' in json_response:
            errors = json_response['errors']
            for error in errors:
                print(f"GraphQL Error: {error.get('message')}")
            raise Exception("GraphQL request failed with errors.")


class BaseApi:
    """ Base class for handling file uploads to Conviso platform """

    def __init__(self, conviso_graphql_client):
        self._conviso_graphql_client = conviso_graphql_client

    def send_file(self, company_id, asset_id, file_path, api_key, mutation):
        """ Generic method for sending a file """
        url = self._conviso_graphql_client.url

        operations = {
            "query": mutation,
            "variables": {
                "companyId": company_id,
                "assetId": asset_id,
                "file": None
            }
        }

        file_map = {"0": ["variables.file"]}

        with open(file_path, 'rb') as file:
            files = {
                'operations': (None, json.dumps(operations), 'application/json'),
                'map': (None, json.dumps(file_map), 'application/json'),
                '0': (file_path, file, 'application/octet-stream')
            }

            headers = {
                'x-api-key': api_key,
                "User-Agent": f"AST:{__version__}"
            }

            response = requests.post(url, files=files, headers=headers)
            response.raise_for_status()
            json_response = response.json()

            self._handle_graphql_errors(json_response)
            return json_response.get('data')

    @staticmethod
    def _handle_graphql_errors(json_response):
        """ Handle GraphQL errors """
        if 'errors' in json_response:
            for error in json_response['errors']:
                print(f"GraphQL Error: {error.get('message')}")
            raise Exception("GraphQL request failed with errors.")


class SbomApi(BaseApi):
    """ Class for SBOM file resources """

    def send_sbom_file(self, company_id, asset_id, file_path, api_key):
        return self.send_file(company_id, asset_id, file_path, api_key, mutations.IMPORT_SBOM)


class ContainerApi(BaseApi):
    """ Class for container finding file resources """

    def send_container_file(self, company_id, asset_id, file_path, api_key):
        return self.send_file(company_id, asset_id, file_path, api_key, mutations.IMPORT_CONTAINER)


class LogAstError(object):
    """ Class to send AST errors to Conviso Platform """
    def __init__(self, conviso_graphql_client):
        self._conviso_graphql_client = conviso_graphql_client

    def send_ast_error(self, company_id, asset_id, log):
        """ send log with company and asset to Conviso Platform """

        graphql_variables = {
            'companyId': str(company_id),
            'assetId': str(asset_id),
            'log': log
        }

        graphql_body_response = self._conviso_graphql_client.execute(
            mutations.LOG_AST_ERROR,
            graphql_variables
        )

        expected_path = 'success'

        success = jmespath.search(
            expected_path,
            graphql_body_response
        )

        return success

class ControlSyncStatus:
    """ Create control sync status on conviso platform """

    def __init__(self, conviso_client):
        self._conviso_graphql_client = conviso_client

    def create_control_sync_status(self, asset_id):
        """ When AST start will create a control sync status using this method """

        graphql_variables = {
            'assetId': str(asset_id)
        }

        graphql_response = self._conviso_graphql_client.execute(
            mutations.CREATE_CONTROL_SYNC_STATUS,
            graphql_variables
        )

        expected_path = 'createControlSyncStatus.controlSyncStatus'

        control_sync_status = jmespath.search(
            expected_path,
            graphql_response
        )

        return control_sync_status

    def update_control_sync_status(self, control_sync_status_id, external_vulnerability_count=1):
        """ This update control sync status by 1 only to set success or failure later """

        graphql_variables = {
            'id': str(control_sync_status_id),
            'externalVulnerabilityCount': external_vulnerability_count
        }

        graphql_response = self._conviso_graphql_client.execute(
            mutations.UPDATE_CONTROL_SYNC_STATUS,
            graphql_variables
        )

        expected_path = 'updateControlSyncStatus.controlSyncStatus'

        control_sync_status = jmespath.search(
            expected_path,
            graphql_response
        )

        return control_sync_status

    def increase_count(self, control_sync_status_id, asset_id, success_count=0, failure_count=0, failure_reason=None):
        """ This is used when scan runs successfully """

        graphql_variables = {
            'controlSyncStatusId': str(control_sync_status_id),
            'assetId': str(asset_id),
            'successCount': success_count,
            'failureCount': failure_count,
            'failureReason': failure_reason
        }

        graphql_response = self._conviso_graphql_client.execute(
            mutations.INCREASE_ISSUE_SCAN_COUNT,
            graphql_variables
        )

        expected_path = 'increaseIssueScanCount.controlSyncStatus'

        result = jmespath.search(
            expected_path,
            graphql_response
        )

        return result



CREATE_SAST_FINDING_INPUT = """
mutation createSastFinding($input: CreateSastFindingInput!) {
  createSastFinding(input: $input) {
    issue {
      id
    }
  }
}
"""

CREATE_SCA_FINDING_INPUT = """
mutation createScaFinding($input: CreateScaFindingInput!) {
    createScaFinding(input: $input) {
        issue {
            id
        }
    }
}
"""

CREATE_IAC_FINDING_INPUT = """
mutation createOrUpdateIacFinding($input: CreateOrUpdateSastFindingInput!) {
    createOrUpdateIacFinding(input: $input) {
        issue {
            id
        }
    }
}
"""

CREATE_CONTAINER_FINDING_INPUT = """
mutation createOrUpdateContainerFinding($input: CreateOrUpdateContainerFindingInput!) {
    createOrUpdateContainerFinding(input: $input) {
        issue {
            id
        }
    }
}
"""

UPDATE_ISSUE_STATUS = """
mutation (
  $issueId: ID!,
  $status: IssueStatusLabel!,
  $reason: String
  $controlSyncStatusId: ID
) {
  changeIssueStatus (
    input: {
      id: $issueId
      status: $status
      reason: $reason
      controlSyncStatusId: $controlSyncStatusId
    }
  ) {
    issue {
      id
    }
  }
}
"""



from datetime import datetime, timezone


class CreateProjectInput:
    def __init__(self, company_id, asset_id, label):
        self.company_id = company_id
        self.asset_id = asset_id
        self.label = label

    def to_graphql_dict(self):
        current_datetime = datetime.now(timezone.utc)
        formatted_date = current_datetime.isoformat()

        return {
            "companyId": int(self.company_id),
            "assetsIds": int(self.asset_id),
            "label": self.label,
            "startDate": str(formatted_date),
            "typeId": 33,
            "playbooksIds": 1,
            "goal": '',
            "scope": self.company_id
        }


class UpdateProjectInput:
    def __init__(self, project_id, asset_id):
        self.project_id = project_id
        self.asset_id = asset_id

    def to_graphql_dict(self):
        return {
            "projectId": int(self.project_id),
            "assetsIds": [int(self.asset_id)],
        }



GET_ASSETS = """ 
query (
  $id: ID!,
  $name: String!,
  $page: Int,
  $limit: Int
) {
  assets(
    companyId: $id
    page: $page
    limit: $limit
    search: {
      name: $name
    }
  ) {
    collection {
      id
      name
      createdAt
      projects(includeAst: true) {
        type
        apiCode
        label
      }
    }

    metadata {
      currentPage
      limitValue
      totalCount
      totalPages
    }
  }
}
"""

GET_ISSUES_STATS = """
query (
  $asset_id: [ID!],
  $company_id: ID!,
  $statuses: [IssueStatusLabel!],
  $end_date: ISO8601DateTime
) {
  issuesStats(
    companyId: $company_id
    filters: {
      assetIds: $asset_id
      statuses: $statuses
      createdAtRange: {
        endDate: $end_date
      }
    }
  ) {
    severities {
      value
      count
    }
  }
}
"""

GET_PROJECTS = """
query (
  $project_code: String!,
  $project_label: String!,
  $company_id: ID!,
  $page: Int,
  $limit: Int
) {
  projects(
    page: $page
    limit: $limit
    params: {
      apiCodeEq: $project_code
      labelEq: $project_label
      scopeIdEq: $company_id
      showHidden: true
      projectTypeLabelEq: "ast"
    }
  ) {
    collection {
      id
      apiCode
      assets {
        id
        name
      }
      company {
        id
        customFeatures
      }
    }
    metadata {
      currentPage
      limitValue
      totalCount
      totalPages
    }
  }
}
"""

GET_COMPANY = """
query get_company($company_id: ID!) {
    company(id: $company_id) {
      id
      label
      customFeatures
    }
}
"""

GET_COMPANIES = """
query Companies {
  companies (
    limit: 50, 
    order: label,
    orderType: ASC
  )  {
    collection {
      id
      label
      customFeatures
    }
  }
}
"""

GET_ISSUES_FINGERPRINT = """
query GetIssuesFingerprint(
    $company_id: ID!,
    $page: Int,
    $per_page: Int,
    $asset_id: [ID!],
    $statuses: [IssueStatusLabel!]
    $failure_types: [Issue!]
) {
    issues(
        companyId: $company_id,
        pagination: {
            page: $page,
            perPage: $per_page
        },
        filters: {
            assetIds: $asset_id,
            statuses: $statuses,
            failureTypes: $failure_types
        }
    ) {
        collection {
            id
            type
            ... on FindingInterface {
                originalIssueIdFromTool
                scanSource
            }
            status
        }
        metadata {
            totalCount
            totalPages
        }
    }
}
"""

GET_DEPLOYS_BY_ASSET = """
query GetDeploysByAsset(
  $asset_id: ID!
) {
  deploysByAsset(
    assetId: $asset_id
  ) {
    collection {
      currentCommit
      previousCommit
    }
  }
}
"""


import copy
import json
from contextlib import suppress
from os import SEEK_SET
from urllib.parse import urljoin

import jsonschema
import requests

PRODUCTION_API_URL = "https://api.convisoappsec.com"
STAGING_API_URL = "https://api.staging.convisoappsec.com"
DEVELOPMENT_API_URL = "http://localhost:3000"
DEFAULT_API_URL = PRODUCTION_API_URL


class RequestsSession(requests.Session):

    def __init__(self, base_url):
        super().__init__()
        self.base_url = base_url

    def request(self, method, url, *args, **kwargs):
        url = urljoin(self.base_url, url)

        return super().request(
            method, url, *args, **kwargs
        )


class FlowAPIException(Exception):
    pass


class FlowAPIAccessDeniedException(FlowAPIException):
    pass


class DeployNotFoundException(FlowAPIException):
    pass


class DockerRegistry(object):
    SAST_ENDPOINT = '/auth/public_auth'

    def __init__(self, client):
        self.client = client

    def get_sast_token(self):
        session = self.client.requests_session
        response = session.get(self.SAST_ENDPOINT)
        response.raise_for_status()
        return response.text


class RESTClient(object):

    def __init__(
        self,
        url=STAGING_API_URL,
        key=None,
        insecure=False,
        user_agent=None,
        ci_provider_name=None
    ):
        self.url = url
        self.insecure = insecure
        self.key = key
        self.user_agent = user_agent
        self.ci_provider_name = ci_provider_name

    @property
    def requests_session(self):
        session = RequestsSession(self.url)
        session.verify = not self.insecure

        session.headers.update({
            'x-api-key': self.key,
            'x-flowcli-ci-provider-name': self.ci_provider_name
        })

        if self.user_agent:
            user_agent_header = {}
            name = self.user_agent.get('name')
            version = self.user_agent.get('version')

            if name and version:
                user_agent_header_fmt = "{name}/{version}"
                user_agent_header_content = user_agent_header_fmt.format(
                    name=name,
                    version=version,
                )

                user_agent_header = {
                    'User-Agent': user_agent_header_content
                }

            session.headers.update(user_agent_header)

        return session


    @property
    def docker_registry(self):
        return DockerRegistry(self)



import click

from convisoappsec.flowcli import help_option

from .values import values
from .tag_tracker import tag_tracker


@click.group('with')
@help_option
def with_():
    pass


with_.add_command(values)
with_.add_command(tag_tracker)

with_.epilog = '''
  Run flow deploy create with COMMAND --help for more information on a command.
'''



import traceback
import click
import json
import subprocess
import shutil
import os
import datetime
from convisoappsec.flowcli import help_option
from convisoappsec.flowcli.context import pass_flow_context
from convisoappsec.logger import log_and_notify_ast_event
from convisoappsec.flowcli.requirements_verifier import RequirementsVerifier
from copy import deepcopy as clone
from convisoappsec.flowcli.common import asset_id_option
from convisoappsec.flowcli.vulnerability.container_vulnerability_manager import ContainerVulnerabilityManager

DEBUG_MODE = False


@click.command()
@asset_id_option(required=False)
@click.option(
    '--debug',
    is_flag=True,
    help='Enable debug mode.'
)
@click.option(
    '-r',
    '--repository-dir',
    default=".",
    show_default=True,
    type=click.Path(
        exists=True,
        resolve_path=True,
    ),
    required=False,
    help="The source code repository directory.",
)
@click.option(
    "--send-to-flow/--no-send-to-flow",
    default=True,
    show_default=True,
    required=False,
    hidden=True,
    help="""Enable or disable the ability of send analysis result
    reports to flow.""",
)
@click.option(
    "--company-id",
    required=False,
    envvar=("CONVISO_COMPANY_ID", "FLOW_COMPANY_ID"),
    help="Company ID on Conviso Platform",
)
@click.option(
    '--asset-name',
    required=False,
    envvar=("CONVISO_ASSET_NAME", "FLOW_ASSET_NAME"),
    help="Provides a asset name.",
)
@click.option(
    '--vulnerability-auto-close',
    default=False,
    is_flag=True,
    hidden=True,
    help="Enable auto fixing vulnerabilities on cp.",
)
@click.argument('image_name')
@help_option
@pass_flow_context
@click.pass_context
def run(
        context, flow_context, asset_id, debug, company_id, repository_dir,
        send_to_flow, asset_name, vulnerability_auto_close, image_name,

):
    """ Run command for container vulnerability scan focused on OS vulnerabilities """
    global DEBUG_MODE
    DEBUG_MODE = debug
    start_time = datetime.datetime.now()

    if send_to_flow:
        prepared_context = RequirementsVerifier.prepare_context(clone(context))

        if debug:
            debug_message(f"Context after being prepared: {prepared_context.params}")

        params_to_copy = [
            'asset_id', 'send_to_flow', 'asset_name', 'vulnerability_auto_close',
            'repository_dir', 'company_id'
        ]

        for param_name in params_to_copy:
            context.params[param_name] = (
                    locals()[param_name] or prepared_context.params[param_name]
            )

        asset_id = context.params['asset_id']
        company_id = context.params['company_id']
    else:
        # this just verify if the api key is valid.
        RequirementsVerifier.list_assets(company_id=company_id, asset_name='example', scan_type='SAST')

        if debug:
            debug_message("User validated!")

    if command_exists('trivy'):
        if debug:
            debug_message("Trivy already installed.")

        scan_command = f"trivy image --pkg-types os --format json --output result.json {image_name}"
    else:
        if debug:
            debug_message("Installing trivy ...")

        subprocess.run(
            "curl -sfL https://raw.githubusercontent.com/aquasecurity/trivy/main/contrib/install.sh | sh -s -- -b conviso/ v0.57.1",
            shell=True,
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL
        )

        if debug:
            debug_message("Trivy has been installed successfully!")

        scan_command = [f"./conviso/trivy image --pkg-types os --format json --output result.json {image_name}"]

    try:
        log_func(f"🔧 Scanning image {image_name} ...")

        if debug:
            debug_message(f"Running the following command: {scan_command}")

        run_command(scan_command)
        log_func("✅ Scan completed successfully.")

        directory = 'conviso/'
        if os.path.isdir(directory):
            if debug:
                debug_message(f"Removing the trivy installation dir, {directory}")
            shutil.rmtree(directory)

        if send_to_flow:
            send_to_conviso_plataform(flow_context, asset_id, company_id)
        else:
            output_results()

        end_time = datetime.datetime.now()

        if debug:
            execution_time = end_time - start_time
            debug_message(f"Total execution time: {execution_time.total_seconds():.2f} seconds.")

        try:
            if vulnerability_auto_close is True:
                vulnerability_manager = ContainerVulnerabilityManager()
                vulnerability_manager.close_vulnerability()
        except Exception:
            log_func("An issue occurred while attempting to fix vulnerabilities. Our technical team has been notified.")
            return

    except Exception as error:
        log_func(f"❌ Scan failed: {error}")


def run_command(command):
    """
    Runs a shell command and logs its execution.

    Args:
        command (str): The scan command to execute.

    Returns:
        The result of a subproccess execution.
    """
    result = subprocess.run(command, shell=True, check=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)

    return result

def send_to_conviso_plataform(flow_context, asset_id, company_id):
    """
    Process and send result to conviso platform.

    This method read the result file, parse and try to send all founded vulnerabilities
    If in any part of the process receive an error, should notify the ast-channel on conviso slack.

    Args:
        conviso_api (object): Responsable to comunicate with conviso graphql api.
        flow_content (object): Some helper methods.
        asset_id (int): The asset where the result will be sended.
        company_id (int): The user company on conviso platform.

    Returns:
        str: Could return a message inform about no vulnerabilities or None
    """
    log_func("🔧 Processing results ...")
    result_file = "result.json"

    try:
        vulnerabilities = extract_vulnerabilities(result_file)

        if vulnerabilities:
            log_func("🔍 Sending vulnerabilities to conviso platform.")
            api_key = flow_context.key
            conviso_api = flow_context.create_conviso_graphql_client()
            conviso_api.container.send_container_file(
                company_id=company_id, asset_id=asset_id, file_path=result_file, api_key=api_key
            )
            log_func("✅ Successfully!")
        else:
            log_func("✅ No vulnerabilities found.")

    except FileNotFoundError:
        log_func(f"❌ {result_file} not found. Ensure the scan was successful.")
        full_trace = traceback.format_exc()
        log_and_notify_ast_event(
            flow_context=flow_context, company_id=company_id, asset_id=asset_id, ast_log=full_trace
        )
    except json.JSONDecodeError:
        log_func(f"❌ Failed to parse {result_file}. Ensure it is valid JSON.")
        full_trace = traceback.format_exc()
        log_and_notify_ast_event(
            flow_context=flow_context, company_id=company_id, asset_id=asset_id, ast_log=full_trace
        )
    except Exception:
        full_trace = traceback.format_exc()
        log_func(f"❌ An error occurred while processing results: {full_trace}")
        log_and_notify_ast_event(
            flow_context=flow_context, company_id=company_id, asset_id=asset_id, ast_log=full_trace
        )


def output_results():
    """
    Output the scan result in case the user don't want to send to conviso platform.
    """
    result_file = "result.json"

    try:
        vulnerabilities = extract_vulnerabilities(result_file)

        if vulnerabilities:
            log_func(f"🔍 Found: {len(vulnerabilities)} vulnerabilities!")
        else:
            log_func("✅ No vulnerabilities found.")

    except Exception:
        full_trace = traceback.format_exc()
        log_func(f"❌ An error occurred while processing results: {full_trace}")


def command_exists(command):
    """
    Validates if a command exists.

    Args: command (str): Command to validate.
    Returns:
        bool: True if the command exists.
    """
    return shutil.which(command) is not None

def extract_vulnerabilities(result_file):
    """Reads a JSON scan result file and extracts vulnerabilities."""
    with open(result_file, 'r') as file:
        scan_results = json.load(file)

    results = scan_results.get("Results", [])
    if results and isinstance(results, list) and len(results) > 0:
        return results[0].get("Vulnerabilities", [])

    return []


def log_func(msg, new_line=True):
    """
    Output a message to the console with styled formatting.

    This function uses `click` to output a styled message to the console. It supports
    controlling whether the message ends with a newline and writes the output to `stderr`.

    Args:
        msg (str): The message to log.
        new_line (bool, optional): Whether to append a newline at the end of the message.
            Defaults to True.

    Returns:
        str: The output of the message.
    """
    click.echo(click.style(msg), nl=new_line, err=True)


def debug_message(msg, new_line=True):
    """
    If debug mode is enabled, this function should be
    used for all debug messages and the message will be styled in orange.
    Otherwise, it uses the default styling.

        Args:
        msg (str): The message to log.
        new_line (bool, optional): Whether to append a newline at the end of the message.
            Defaults to True.

    Returns:
        str: The output of the message.
    """
    timestamp = datetime.datetime.now().strftime("%Y-%m-%d %H:%M:%S")
    style = {"fg": "bright_yellow"} if DEBUG_MODE else {}
    click.echo(click.style(f"🪲 [{timestamp}] DEBUG: {msg}", **style), nl=new_line, err=True)



import click
import click_log
import json
from base64 import b64decode
from re import search as regex_search
from copy import deepcopy as clone
from convisoappsec.flowcli import help_option
from convisoappsec.flowcli.context import pass_flow_context
from convisoappsec.logger import LOGGER
from convisoappsec.flowcli.requirements_verifier import RequirementsVerifier
from convisoappsec.flow.graphql_api.beta.models.issues.sast import (CreateSastFindingInput)
from convisoappsec.flow.graphql_api.beta.models.issues.sca import CreateScaFindingInput
from convisoappsec.common.graphql.errors import ResponseError

click_log.basic_config(LOGGER)


@click.command()
@click.option(
    "-i",
    "--input-file",
    required=True,
    type=click.Path(exists=True),
    help='The path to SARIF file.',
)
@click.option(
    "--company-id",
    required=False,
    envvar=("CONVISO_COMPANY_ID", "FLOW_COMPANY_ID"),
    help="Company ID on Conviso Platform",
)
@click.option(
    "-r",
    "--repository-dir",
    default=".",
    show_default=True,
    type=click.Path(
        exists=True,
        resolve_path=True,
    ),
    required=False,
    help="The source code repository directory.",
)
@click.option(
    '--asset-name',
    required=False,
    envvar=("CONVISO_ASSET_NAME", "FLOW_ASSET_NAME"),
    help="Provides a asset name.",
)
@help_option
@pass_flow_context
@click.pass_context
def import_sarif(context, flow_context, input_file, company_id, repository_dir, asset_name):
    context.params['company_id'] = company_id if company_id is not None else None
    context.params['repository_dir'] = repository_dir

    prepared_context = RequirementsVerifier.prepare_context(clone(context))
    asset_id = prepared_context.params['asset_id']

    try:
        conviso_api = flow_context.create_conviso_api_client_beta()
        LOGGER.info("💬 Starting the import process for the SARIF file.")
        parse_sarif_file(conviso_api, asset_id, input_file)
    except Exception as e:
        LOGGER.error(f"❌ Error during SARIF file import: {str(e)}")
        raise Exception("SARIF file import failed. Please contact support and provide the SARIF file for assistance.")


def parse_sarif_file(conviso_api, asset_id, sarif_file):
    try:
        sarif_data = _load_sarif_data(sarif_file)
        if not sarif_data:
            raise Exception("Failed to load SARIF file")

        sarif_infos = _extract_rules_info(sarif_data)

        success_count = 0
        error_count = 0

        for run in sarif_data.get('runs', []):
            for result in run.get('results', []):
                try:
                    if _process_vulnerability(conviso_api, asset_id, result, sarif_infos):
                        success_count += 1
                    else:
                        error_count += 1
                except Exception as e:
                    LOGGER.error(f"❌ Error processing vulnerability: {str(e)}")
                    error_count += 1

        LOGGER.info(f"✅ Successfully processed {success_count} vulnerabilities")
        if error_count > 0:
            LOGGER.warning(f"⚠️ {error_count} vulnerabilities with errors")

        LOGGER.info("✅ SARIF file import completed successfully.")

    except Exception as e:
        LOGGER.error(f"❌ Error during SARIF file parsing: {str(e)}")
        raise


def _load_sarif_data(sarif_file):
    try:
        cleaned_file = clean_file(sarif_file)

        with open(cleaned_file, 'r', encoding='utf-8') as file:
            sarif_data = json.load(file)

        if not isinstance(sarif_data, dict) or 'runs' not in sarif_data:
            LOGGER.error("❌ Invalid SARIF file: 'runs' structure not found")
            return None

        return sarif_data

    except json.JSONDecodeError as e:
        LOGGER.error(f"❌ Error decoding JSON: {str(e)}")
        return None
    except Exception as e:
        LOGGER.error(f"❌ Error loading file: {str(e)}")
        return None


def _extract_rules_info(sarif_data):
    sarif_infos = []

    for run in sarif_data.get('runs', []):
        for rule in run.get('tool', {}).get('driver', {}).get('rules', []):
            title = (rule.get('shortDescription', {}).get('text') or
                     rule.get('name') or
                     rule.get('fullDescription', {}).get('text') or
                     rule.get('help', {}).get('text'))

            rule_info = {
                "id": rule.get('id'),
                "name": title,
                "references": rule.get('helpUri'),
                "description": _get_rule_description(rule)
            }
            sarif_infos.append(rule_info)

    return sarif_infos


def _get_rule_description(rule):
    return (rule.get('help', {}).get('text') or
            rule.get('fullDescription', {}).get('text') or
            rule.get('shortDescription', {}).get('text'))


def _process_vulnerability(conviso_api, asset_id, result, sarif_infos):
    try:
        if not result.get('ruleId'):
            LOGGER.warning("⚠️ Result without ruleId, skipping...")
            return False

        vuln_data = _extract_vulnerability_data(result, sarif_infos)
        if not vuln_data:
            LOGGER.warning(f"⚠️ Could not extract data for {result.get('ruleId')}")
            return False

        rule_id = result.get('ruleId', '')

        if "(sca)" in rule_id.lower():
            return _process_sca_vulnerability(conviso_api, asset_id, result, vuln_data)
        elif "(sast)" in rule_id.lower():
            return _process_sast_vulnerability(conviso_api, asset_id, result, vuln_data)
        else:
            return _process_sast_vulnerability(conviso_api, asset_id, result, vuln_data)

    except Exception as e:
        LOGGER.error(f"❌ Error processing vulnerability: {str(e)}")
        return False


def _extract_vulnerability_data(result, sarif_infos):
    try:
        rule_id = result.get('ruleId')
        matching_info = next((info for info in sarif_infos if info['id'] == rule_id), None)
        title = None
        references = None
        description = None

        if matching_info:
            title = matching_info['name']
            references = matching_info['references']
            description = matching_info['description']

        if not title:
            title = result.get('message', {}).get('text', 'No title provided')
        if not description:
            description = result.get('message', {}).get('text', 'No description provided')

        locations = result.get('locations', [])
        if not locations:
            LOGGER.warning(f"⚠️ No location found for {rule_id}")
            return None

        first_location = locations[0]
        physical_location = first_location.get('physicalLocation', {})

        return {
            'title': title,
            'references': references,
            'description': description,
            'severity': result.get('level', 'info'),
            'file_name': physical_location.get('artifactLocation', {}).get('uri'),
            'vulnerable_line': physical_location.get('region', {}).get('startLine'),
            'first_line': physical_location.get('region', {}).get('startLine', 1),
            'code_snippet': physical_location.get('contextRegion', {}).get('snippet', {}).get('text', ''),
        }

    except Exception as e:
        LOGGER.error(f"❌ Error extracting vulnerability data: {str(e)}")
        return None


def _process_sca_vulnerability(conviso_api, asset_id, result, vuln_data):
    try:
        message_text = result.get('message', {}).get('text', '')
        sca_data = _parse_sca_data(message_text)

        if not sca_data:
            LOGGER.warning(f"⚠️ Could not extract SCA data from: {message_text}")
            return False

        create_sca_vulnerabilities(
            conviso_api, asset_id, message_text,
            vuln_data['references'], vuln_data['description'],
            vuln_data['severity'], vuln_data['file_name'],
            vuln_data['first_line'], sca_data['package'],
            sca_data['version'], sca_data['cve']
        )

        return True

    except Exception as e:
        LOGGER.error(f"❌ Error processing SCA vulnerability: {str(e)}")
        return False


def _parse_sca_data(message_text):
    try:
        if ':' not in message_text:
            return None

        package = message_text.split(':')[1].split(' ')[0]
        version = package.split('-')[-1] if '-' in package else 'Unknown'
        cve = 'Unknown'

        if '(' in message_text and ')' in message_text:
            parts = message_text.split(' ')
            for part in parts:
                if part.startswith('(') and part.endswith(')'):
                    cve = part.strip('()')
                    break

        return {
            'package': package,
            'version': version,
            'cve': cve
        }

    except Exception as e:
        LOGGER.error(f"❌ Error parsing SCA data: {str(e)}")
        return None


def _process_sast_vulnerability(conviso_api, asset_id, result, vuln_data):
    try:
        create_sast_vulnerabilities(
            conviso_api, asset_id, vuln_data['title'],
            vuln_data['references'], vuln_data['description'],
            vuln_data['vulnerable_line'], vuln_data['severity'],
            vuln_data['file_name'], vuln_data['code_snippet'],
            vuln_data['first_line'], None
        )

        return True

    except Exception as e:
        LOGGER.error(f"❌ Error processing SAST vulnerability: {str(e)}")
        return False


def create_sast_vulnerabilities(conviso_api, asset_id, *args):
    title, references, description, vulnerable_line, severity, file_name, code_snippet, first_line, cve = args

    issue_model = CreateSastFindingInput(
        asset_id=asset_id,
        file_name=file_name,
        vulnerable_line=vulnerable_line or 0,
        title=title or 'No title provided',
        description=description or 'No description provided',
        severity=severity,
        commit_ref=None,
        deploy_id=None,
        code_snippet=parse_code_snippet(code_snippet),
        reference=parse_conviso_references(references),
        first_line=first_line,
        category=None,
        original_issue_id_from_tool=None,
        solution=None
    )

    try:
        conviso_api.issues.create_sast(issue_model)
        LOGGER.debug(f"✅ SAST vulnerability created: {title}")
    except ResponseError as error:
        if error.code == 'RECORD_NOT_UNIQUE':
            LOGGER.debug(f"ℹ️ SAST vulnerability already exists: {title}")
        else:
            LOGGER.error(f"❌ Error creating SAST vulnerability: {error}")
            raise
    except Exception as e:
        LOGGER.error(f"❌ Unexpected error creating SAST vulnerability: {str(e)}")
        raise


def create_sca_vulnerabilities(conviso_api, asset_id, *args):
    title, references, description, severity, file_name, first_line, package, version, cve = args

    issue_model = CreateScaFindingInput(
        asset_id=asset_id,
        title=title,
        description=description,
        severity=severity,
        solution="Update to the last package version.",
        reference=references,
        file_name=file_name,
        affected_version=version,
        package=package,
        cve=cve,
        patched_version='',
        category='',
        original_issue_id_from_tool=''
    )

    try:
        conviso_api.issues.create_sca(issue_model)
        LOGGER.debug(f"✅ SCA vulnerability created: {title}")
    except ResponseError as error:
        if error.code == 'RECORD_NOT_UNIQUE':
            LOGGER.debug(f"ℹ️ SCA vulnerability already exists: {title}")
        else:
            LOGGER.error(f"❌ Error creating SCA vulnerability: {error}")
            raise
    except Exception as e:
        LOGGER.error(f"❌ Unexpected error creating SCA vulnerability: {str(e)}")
        raise


def parse_code_snippet(code_snippet):
    try:
        decoded_text = b64decode(code_snippet).decode("utf-8")
        lines = decoded_text.split("\n")
        cleaned_lines = []

        for line in lines:
            cleaned_line = line.split(": ", 1)[-1]
            cleaned_lines.append(cleaned_line)

        code_snippet = "\n".join(cleaned_lines)

        return code_snippet
    except Exception:
        return code_snippet


def parse_conviso_references(references=[]):
    if not references:
        return ""

    DIVIDER = "\n"

    references_to_join = []

    for reference in references:
        if reference:
            references_to_join.append(reference)

    return DIVIDER.join(references_to_join)


def parse_first_line_number(encoded_base64):
    decoded_text = b64decode(encoded_base64).decode("utf-8")

    regex = r"^(\d+):"

    result = regex_search(regex, decoded_text)

    if result and result.group(1):
        return result.group(1)

    LINE_NUMBER_WHEN_NOT_FOUND = 1
    return LINE_NUMBER_WHEN_NOT_FOUND


def clean_file(input_file):
    with open(input_file, mode="rb") as file:
        content = file.read()

    if content.startswith(b'\xef\xbb\xbf'):
        content = content[3:]

    cleaned_file = input_file + ".cleaned"
    with open(cleaned_file, mode="wb") as file:
        file.write(content)

    return cleaned_file


import_sarif.epilog = '''
'''
EPILOG = '''
Examples:

  \b
  1 - Import results on SARIF file to Conviso Platform:
    $ export CONVISO_API_KEY='your-api-key'
    $ {command} --input-file /path/to/file.sarif

'''  # noqa: E501

SHORT_HELP = "Perform import of vulnerabilities from SARIF file to Conviso Platform"

command = 'conviso findings import-sarif'
import_sarif.short_help = SHORT_HELP
import_sarif.epilog = EPILOG.format(
    command=command,
)



import yaml
import tempfile

from .source_code_scanner import SourceCodeScanner
from .exceptions import SourceCodeScannerException


class SCC(SourceCodeScanner):

    def __init__(
        self, 
        source_code_dir, 
        container_source_dir = '/code', 
        create_source_code_volume = True
    ):
        super().__init__(
            source_code_dir, 
            create_source_code_volume=create_source_code_volume
        )
        self.__scan_result = {}
        self.__container_source_dir = container_source_dir

    @property
    def repository(self):
        return 'convisoappsec/scc'

    @property
    def tag(self):
        return 'latest'

    @property
    def container_source_dir(self):
        return self.__container_source_dir

    def _read_scan_stdout(self, stdout_generator):
        with tempfile.TemporaryFile() as yaml_output:
            for chunk in stdout_generator:
                yaml_output.write(chunk)

            yaml_output.seek(0)

            self.__scan_result = yaml.load(
                yaml_output,
                Loader=yaml.FullLoader
            )

    @property
    def summary(self):
        summary = self.__scan_result.get('SUM')
        if not summary:
            raise SourceCodeScannerException(
                'Unexpected error retrienving source code summary metrics'
            )

        return summary

    @property
    def total_source_code_lines(self):
        return self.summary.get('code')

    @property
    def command(self):
        return [
            '--no-cocomo',
            '--no-complexity',
            '--format',
            'cloc-yaml'
        ]



import json
import click
import click_log
import traceback
from convisoappsec.common.retry_handler import RetryHandler
from convisoappsec.common.box import ContainerWrapper
from convisoappsec.common import strings
from convisoappsec.flow.graphql_api.beta.models.issues.sca import CreateScaFindingInput
from convisoappsec.flowcli import help_option
from convisoappsec.flowcli.common import asset_id_option, on_http_error
from convisoappsec.flowcli.context import pass_flow_context
from convisoappsec.logger import LOGGER, log_and_notify_ast_event
from convisoappsec.common.graphql.errors import ResponseError
from convisoappsec.flowcli.requirements_verifier import RequirementsVerifier
from copy import deepcopy as clone
from convisoappsec.flowcli.sbom import sbom
from convisoappsec.flowcli.vulnerability.run import perform_sca_scan, close_sca_issues, reopen_issues
from convisoappsec.common.cleaner import Cleaner

click_log.basic_config(LOGGER)


@click.command()
@click_log.simple_verbosity_option(LOGGER)
@asset_id_option(required=False)
@click.option(
    '-r',
    '--repository-dir',
    default=".",
    show_default=True,
    type=click.Path(
        exists=True,
        resolve_path=True,
    ),
    required=False,
    help="The source code repository directory.",
)
@click.option(
    "--send-to-flow/--no-send-to-flow",
    default=True,
    show_default=True,
    required=False,
    help="""Enable or disable the ability of send analysis result
    reports to flow.""",
    hidden=True
)
@click.option(
    "--custom-sca-tags",
    hidden=True,
    required=False,
    multiple=True,
    type=(str, str),
    help="""It should be passed as <repository_name> <image_tag>. It accepts multiple values"""
)
@click.option(
    "--scanner-timeout",
    hidden=True,
    required=False,
    default=7200,
    type=int,
    help="Set timeout for each scanner"
)
@click.option(
    "--parallel-workers",
    hidden=True,
    required=False,
    default=2,
    type=int,
    help="Set max parallel workers"
)
@click.option(
    "--deploy-id",
    default=None,
    required=False,
    hidden=True,
    envvar=("CONVISO_DEPLOY_ID", "FLOW_DEPLOY_ID")
)
@click.option(
    '--experimental',
    default=False,
    is_flag=True,
    hidden=True,
    help="Enable experimental features.",
)
@click.option(
    "--company-id",
    required=False,
    envvar=("CONVISO_COMPANY_ID", "FLOW_COMPANY_ID"),
    help="Company ID on Conviso Platform",
)
@click.option(
    '--asset-name',
    required=False,
    envvar=("CONVISO_ASSET_NAME", "FLOW_ASSET_NAME"),
    help="Provides a asset name.",
)
@click.option(
    '--vulnerability-auto-close',
    default=False,
    is_flag=True,
    hidden=True,
    help="Enable auto fixing vulnerabilities on cp.",
)
@click.option(
    '--from-ast',
    default=False,
    is_flag=True,
    hidden=True,
    help="Internal use only.",
)
@click.option(
    '--cleanup',
    default=False,
    is_flag=True,
    show_default=True,
    help="Clean up system resources, including temporary files, stopped containers, unused Docker images and volumes.",
)
@click.option(
    '--control-sync-status-id',
    required=False,
    hidden=True,
    help="Control sync status id.",
)
@help_option
@pass_flow_context
@click.pass_context
def run(context, flow_context, asset_id, company_id, repository_dir, send_to_flow, custom_sca_tags,
        scanner_timeout, parallel_workers, deploy_id, experimental, asset_name, vulnerability_auto_close, from_ast,
         cleanup, control_sync_status_id):
    """
      This command will perform SCA analysis at the source code. The analysis
      results can be reported or not to flow application.
    """
    if not from_ast:
        prepared_context = RequirementsVerifier.prepare_context(clone(context))

        params_to_copy = [
            'asset_id', 'company_id', 'repository_dir', 'send_to_flow',
            'deploy_id', 'custom_sca_tags', 'scanner_timeout', 'parallel_workers',
            'experimental', 'asset_name', 'vulnerability_auto_close', 'cleanup', 'control_sync_status_id'
        ]

        for param_name in params_to_copy:
            context.params[param_name] = (
                locals()[param_name] or prepared_context.params[param_name]
            )

    perform_command(
        flow_context,
        context,
        context.params['asset_id'],
        context.params['company_id'],
        context.params['repository_dir'],
        context.params['send_to_flow'],
        context.params['custom_sca_tags'],
        context.params['scanner_timeout'],
        context.params['deploy_id'],
        context.params['experimental'],
        from_ast,
        context.params['cleanup'],
        context.params['control_sync_status_id']
    )

def deploy_results_to_conviso(flow_context, conviso_api, results_filepaths, asset_id, company_id, control_sync_status_id):
    """Send vulnerabilities to Conviso platform via GraphQL endpoint."""

    results_context = click.progressbar(
        results_filepaths, label="Sending SCA reports to the Conviso Platform..."
    )

    duplicated_issues = 0
    total_issues = 0

    with results_context as reports:
        for report_path in reports:
            try:
                with open(report_path, 'r') as report_file:
                    report_content = json.load(report_file)

                issues = report_content.get("issues", [])

                for issue in issues:
                    if not issue:
                        continue

                    total_issues += 1
                    description = issue.get("description", "")
                    hash_issue = issue.get('hash_issue', [])
                    cves = next(([item] for item in issue.get("cve", []) if item.startswith("CVE")), [])
                    path = issue.get("path", "")
                    fixed_version = issue.get('fixed_version', {})
                    patched_version = fixed_version.get('fixed') if fixed_version else None
                    description = description or ""
                    sanitezed_description = strings.parse_to_ascii(description)
                    severity = define_severity(issue.get("severity", ""))

                    issue_model = CreateScaFindingInput(
                        asset_id=asset_id,
                        title=issue.get("title", ""),
                        description=sanitezed_description,
                        severity=severity,
                        solution=issue.get("solution", ""),
                        reference=parse_conviso_references(issue.get("references", [])),
                        file_name=get_relative_path(path),
                        affected_version=issue.get("version", "Unknown"),
                        package=issue.get("component", "Unknown"),
                        cve=cves,
                        patched_version=patched_version,
                        category=issue.get('cwe', ''),
                        original_issue_id_from_tool=hash_issue,
                        control_sync_status_id=control_sync_status_id
                    )

                    try:
                        conviso_api.issues.create_sca(issue_model)
                    except ResponseError as error:
                        if error.code == 'RECORD_NOT_UNIQUE':
                            duplicated_issues += 1
                        else:
                            retry_handler = RetryHandler(
                                flow_context=flow_context, company_id=company_id, asset_id=asset_id
                            )
                            retry_handler.execute_with_retry(conviso_api.issues.create_sca, issue_model)

                    except Exception:
                        retry_handler = RetryHandler(
                            flow_context=flow_context, company_id=company_id, asset_id=asset_id
                        )
                        retry_handler.execute_with_retry(conviso_api.issues.create_sca, issue_model)

                    continue

            except (OSError, json.JSONDecodeError):
                LOGGER.warn(f"⚠️ Failed to process the report. Our technical team has been notified.")
                full_trace = traceback.format_exc()
                log_and_notify_ast_event(
                    flow_context=flow_context, company_id=company_id, asset_id=asset_id,
                    ast_log=str(full_trace)
                )
                continue

    LOGGER.info(f"💬 {duplicated_issues} issue(s) ignored due to duplication.")

def define_severity(osv_severity):
    """Map OSV severity levels to Conviso platform severity levels."""
    mapping = {
        "LOW": "LOW",
        "MODERATE": "MEDIUM",
        "HIGH": "HIGH",
        "CRITICAL": "CRITICAL",
    }

    return mapping.get(osv_severity.upper(), "LOW")

def parse_conviso_references(references=[]):
    DIVIDER = "\n"
    urls = [ref['url'] for ref in references]
    return DIVIDER.join(urls)


def get_relative_path(path):
    """
    Returns the full path if the file is in a subdirectory or just the file name if it's in the root directory,
    disregarding the '/code/' prefix.

    :param path: The file path.
    :return: The processed path.
    """

    if not path:
        return ''

    if path.startswith('/code/'):
        relative_path = path[len('/code/'):]
    else:
        relative_path = path

    if '/' in relative_path:
        return relative_path
    else:
        return relative_path.split('/')[-1]


def perform_command(
        flow_context, context, asset_id, company_id, repository_dir, send_to_flow, custom_sca_tags, scanner_timeout,
        deploy_id, experimental, from_ast, cleanup, control_sync_status_id
):
    if send_to_flow and experimental and not asset_id:
        raise click.MissingParameter(
            "It is required when sending reports to Conviso Platform using experimental API.",
            param_type="option",
            param_hint="--asset-id",
        )

    try:
        REQUIRED_CODEBASE_PATH = '/code'
        OSV_SCANNER_IMAGE_NAME = 'osv_scanner'

        scanners = {
            OSV_SCANNER_IMAGE_NAME: {
                'repository_name': OSV_SCANNER_IMAGE_NAME,
                'tag': 'latest',
                'command': [
                    '-c', REQUIRED_CODEBASE_PATH,
                    '-f', 'json',
                    '-o', '/{}.json'.format(OSV_SCANNER_IMAGE_NAME)
                ],
                'repository_dir': repository_dir
            },
        }

        if custom_sca_tags:
            for custom_tag in custom_sca_tags:
                scan_name, tag = custom_tag
                if scan_name in scanners.keys():
                    scanners[scan_name]['tag'] = tag
                else:
                    raise click.BadOptionUsage(
                        option_name='--custom-sca-tags',
                        message="Custom scan {0} or tag {1} invalid".format(
                            scan_name, tag)
                    )

        conviso_rest_api = flow_context.create_conviso_rest_api_client()
        token = conviso_rest_api.docker_registry.get_sast_token()
        LOGGER.info('💬 Preparing Environment...')
        scabox = ContainerWrapper(
            token=token,
            containers_map=scanners,
            logger=LOGGER,
            timeout=scanner_timeout
        )
        LOGGER.info('💬 Starting SCA...')
        scabox.run()

        LOGGER.info('💬 Processing Results...')
        results_filepaths = []
        for unit in scabox.scanners:
            file_path = unit.results
            if file_path:
                results_filepaths.append(file_path)

        if send_to_flow:
            LOGGER.info("Sending data to the Conviso Platform...")
            conviso_beta_api = flow_context.create_conviso_api_client_beta()

            deploy_results_to_conviso(flow_context, conviso_beta_api, results_filepaths, asset_id, company_id, control_sync_status_id)

        # TODO add CI Decision block code
        LOGGER.info('✅ SCA Scan Finished.')

        # Generate SBOM when execute a sca only scan.
        sbom_generate = sbom.commands.get('generate')
        context.params.pop('cleanup', None)
        specific_param = {"from_ast": True}
        context.params.update(specific_param)
        sbom_generate.invoke(context)

        # run auto close for sca run
        if context.params['vulnerability_auto_close'] is True and from_ast == False:
            log_func("[*] Verifying if any vulnerability was fixed...")
            try:
                perform_sca_auto_close(flow_context, company_id, asset_id, repository_dir)
            except Exception:
                LOGGER.warn(f"⚠️ Failed to execute vulnerability auto close. Our technical team has been notified.")
                full_trace = traceback.format_exc()
                log_and_notify_ast_event(
                    flow_context=flow_context, company_id=company_id, asset_id=asset_id,
                    ast_log=str(full_trace)
                )

        if cleanup and from_ast == False:
            LOGGER.info("🧹 Cleaning up ...")
            cleaner = Cleaner()
            cleaner.cleanup()

        if not results_filepaths:
            context.params['sca_vulnerability_count'] = 0
            return

        context.params['sca_vulnerability_count'] = total_vulnerability_count(results_filepaths[0])

    except Exception as e:
        on_http_error(e)
        raise click.ClickException(str(e)) from e


def total_vulnerability_count(file_path: str) -> int:
    """
    Extract the total vulnerability count from a sca scan result file.

    Args:
        file_path (str): Path to JSON result file containing vulnerability scan results.
                        The file should have a 'summary' section with
                        'issues_count.total' field.

    Returns:
        int: Total number of vulnerabilities found in the scan result file.
             Returns 0 if the file doesn't exist or no vulnerabilities are found.
    """

    try:
        with open(file_path, 'r', encoding='utf-8') as f:
            data = json.load(f)
            return len(data.get('issues', []))

    except (FileNotFoundError, KeyError, json.JSONDecodeError):
        return 0


def perform_sca_auto_close(flow_context, company_id, asset_id, repository_dir):
    """ This method perform auto close vulnerabilities for sca only """
    conviso_beta_api = flow_context.create_conviso_api_client_beta()
    statuses = ['CREATED', 'IDENTIFIED', 'IN_PROGRESS', 'AWAITING_VALIDATION', 'FIX_ACCEPTED']
    page = 1
    merged_issues_sca = []

    # get vulnerabilities until last page
    while True:
        issues_from_cp = conviso_beta_api.issues.auto_close_vulnerabilities(
            company_id, asset_id, statuses, page, vulnerability_type='SCA_FINDING'
        )

        total_pages = issues_from_cp['metadata']['totalPages']
        issues_collection = issues_from_cp['collection']
        issues_collection = [item for item in issues_collection if item['scanSource'] == 'conviso_scanner']

        merged_issues_sca.extend(issues_collection)

        if total_pages == page:
            break
        else:
            page += 1

    sca_issues_with_fix_accepted = [item for item in merged_issues_sca if item['status'] == 'FIX_ACCEPTED']
    sca_issues_without_fix_accepted = [item for item in merged_issues_sca if item['status'] != 'FIX_ACCEPTED']

    if len(issues_from_cp) == 0:
        log_func("No vulnerabilities were found on the Conviso Platform!")
        return

    sca_hash_issues = perform_sca_scan(repository_dir=repository_dir)

    set_of_sca_hash_issues = set(sca_hash_issues)
    close_sca_issues(conviso_beta_api, sca_issues_without_fix_accepted, set_of_sca_hash_issues)

    sca_issues_to_reopen = [
        {'id': item['id'], 'originalIssueIdFromTool': item['originalIssueIdFromTool']}
        for item in sca_issues_with_fix_accepted if item['originalIssueIdFromTool'] in sca_hash_issues
    ]

    if sca_issues_to_reopen:
        log_func("SCA: reopening {issues} vulnerability/vulnerabilities on conviso platform ...".format(
            issues=len(sca_issues_to_reopen))
        )

        reopen_issues(conviso_beta_api, sca_issues_to_reopen)

def log_func(msg, new_line=True):
    click.echo(click.style(msg, bold=True, fg='blue'), nl=new_line, err=True)

EPILOG = '''
Examples:

  \b
  1 - Reporting the results to flow api:
    1.1 - Running an analysis at all commit range:
      $ export CONVISO_API_KEY='your-api-key'
      $ {command}

'''  # noqa: E501

SHORT_HELP = "Perform Source Composition analysis"

command = 'conviso sca run'
run.short_help = SHORT_HELP
run.epilog = EPILOG.format(
    command=command,
)



from setuptools import setup, find_packages
from os import path

HERE = path.abspath(
    path.dirname(__file__)
)

ROOT_PACKAGE = 'convisoappsec'
SCRIPTS_SHELL_COMPLETER_DIR = path.join('scripts', 'shell_completer')
VERSION_MODULE = path.join(ROOT_PACKAGE, 'version.py')
README_PATH = path.join(HERE, 'README.md')

version_module_context = {}


with open(VERSION_MODULE) as fp:
    exec(fp.read(), version_module_context)

with open(README_PATH, "r") as fh:
    long_description = fh.read()

version = version_module_context.get('__version__')

setup(
    name='conviso-ast',
    version=version,
    long_description=long_description,
    long_description_content_type='text/markdown',
    maintainer='Conviso',
    maintainer_email='development@convisoappsec.com',
    package_data={'convisoappsec': ['flowcli/vulnerability/rules_schema.json']},
    packages=find_packages(
        exclude=["test*"],
    ),
    install_requires=[
        "GitPython==3.1.45",
        "click==8.1.8",
        "requests==2.32.5",
        "urllib3==2.4.0",
        "semantic-version==2.10.0",
        "docker==7.1.0",
        "PyYAML==6.0.3",
        "click-log==0.4.0",
        "transitions==0.9.2",
        "jsonschema==4.25.1",
        "giturlparse<=0.12.0",
        "jmespath==1.0.1",
        "setuptools==78.1.0"
    ],
    entry_points={
        'console_scripts': [
            'flow=convisoappsec.flowcli.entrypoint:cli',
            'conviso=convisoappsec.flowcli.entrypoint:cli',
        ]
    },
    scripts=[
        path.join(SCRIPTS_SHELL_COMPLETER_DIR, 'flow_bash_completer.sh'),
        path.join(SCRIPTS_SHELL_COMPLETER_DIR, 'flow_zsh_completer.sh'),
        path.join(SCRIPTS_SHELL_COMPLETER_DIR, 'flow_fish_completer.fish'),
    ],
    project_urls={
        'Source': 'https://github.com/convisoappsec/convisocli/',
    },
    python_requires='>=3.9',
)



import semantic_version
import re


class Version(semantic_version.Version):
    PREFIXED_VERSION_RE_FMT = '^{prefix}(.+)'

    def __init__(self, version_string, prefix=None):
        super().__init__(
            self.__remove_prefix(prefix, version_string)
        )

        self.__version_string = version_string

    def __str__(self):
        return self.__version_string

    @classmethod
    def __remove_prefix(cls, prefix, version_string):
        if not prefix:
            return version_string

        version_re = cls.PREFIXED_VERSION_RE_FMT.format(
            prefix=re.escape(prefix)
        )

        match = re.match(version_re, version_string)

        if match:
            return match.group(1)

        return version_string

    def find_previous(self, versions):
        previous_tags = list(
            filter(lambda v: v < self, versions)
        )

        return self.find_latest(previous_tags)

    @staticmethod
    def find_latest(versions):
        if len(versions) > 0:
            return max(versions)




from convisoappsec.common.graphql.low_client import GraphQLClient
from convisoappsec.flow.graphql_api.beta.resources_api import IssuesAPI


class ConvisoGraphQLClientBeta():
    DEFAULT_AUTHORIZATION_HEADER_NAME = 'x-api-key'

    def __init__(self, api_url, api_key):
        headers = {
            self.DEFAULT_AUTHORIZATION_HEADER_NAME: api_key
        }

        self.__low_client = GraphQLClient(api_url, headers)

    @property
    def issues(self):
        return IssuesAPI(self.__low_client)



import warnings

from convisoappsec.flow.versioning_style import semantic_versioning

from .version_searcher_result import VersionSearcherResult


class SortedByVersioningStyle(object):
    SEMANTIC_VERSIONING_STYLE = 'semantic-versioning'
    STYLES = [
        SEMANTIC_VERSIONING_STYLE,
    ]

    def __init__(
        self, version_control_system_adapter,
        ignore_prefix, style, current_tag, **kargs
    ):
        self.version_control_system_adapter = version_control_system_adapter
        self.ignore_prefix = ignore_prefix
        self.style = style
        self.current_tag = current_tag
        self.suppress_warnings = kargs.get('suppress_warnings', True)

    def find_current_and_previous_version(self):
        tags = self.version_control_system_adapter.tags()

        versions = []

        for tag in tags:
            try:
                versions.append(
                    semantic_versioning.Version(
                      tag,
                      prefix=self.ignore_prefix,
                    )
                )
            except ValueError as e:
                if not self.suppress_warnings:
                    warnings.warn(str(e))

        current_version = None

        if not self.current_tag:
            current_version = semantic_versioning.Version.find_latest(
                versions
            )

            if not current_version:
                raise RuntimeError("No current_version was found")
        else:
            current_version = semantic_versioning.Version(
                self.current_tag,
                prefix=self.ignore_prefix,
            )

            if current_version not in versions:
                raise ValueError(
                    "Current version[%s] not exists on repository"
                    %
                    current_version
                )

        previous_version = current_version.find_previous(versions)

        current_tag = str(current_version)
        current_commit = self.version_control_system_adapter.show_commit_from_tag(  # noqa: E501
            current_tag
        )

        previous_tag = None
        previous_commit = self.version_control_system_adapter.empty_repository_tree_commit  # noqa: E501

        if previous_version:
            previous_tag = str(previous_version)
            previous_commit = self.version_control_system_adapter.show_commit_from_tag(  # noqa: E501
                previous_tag
            )

        return VersionSearcherResult(
            current_commit=current_commit,
            previous_commit=previous_commit,
            current_tag=current_tag,
            previous_tag=previous_tag,

        )



from convisoappsec.common.graphql.low_client import GraphQLClient
from convisoappsec.flow.graphql_api.v1.resources_api import AssetsAPI, CompaniesApi, IssuesApi, DeploysApi, \
    SbomApi, LogAstError, ContainerApi, ControlSyncStatus


class ConvisoGraphQLClient():
    DEFAULT_AUTHORIZATION_HEADER_NAME = 'x-api-key'

    def __init__(self, api_url, api_key):
        headers = {
            self.DEFAULT_AUTHORIZATION_HEADER_NAME: api_key
        }

        self.__low_client = GraphQLClient(api_url, headers)

    @property
    def assets(self):
        return AssetsAPI(self.__low_client)

    @property
    def companies(self):
        return CompaniesApi(self.__low_client)

    @property
    def issues(self):
        return IssuesApi(self.__low_client)

    @property
    def deploys(self):
        return DeploysApi(self.__low_client)

    @property
    def sbom(self):
        return SbomApi(self.__low_client)

    @property
    def container(self):
        return ContainerApi(self.__low_client)

    @property
    def ast_errors(self):
        return LogAstError(self.__low_client)

    @property
    def control_sync_status(self):
        return ControlSyncStatus(self.__low_client)



import click
# TODO: refactoring. all deploy create share some behavior
from convisoappsec.flow.util import project_metrics
from convisoappsec.flow.version_searchers import TimeBasedVersionSearcher
from convisoappsec.flow.version_control_system_adapter import GitAdapter
from convisoappsec.flowcli.context import pass_flow_context
from convisoappsec.flowcli import help_option
from convisoappsec.flowcli.common import on_http_error
from convisoappsec.flowcli.deploy.create.context import pass_create_context
from convisoappsec.flowcli.deploy.create.with_.tag_tracker.context import (
    pass_tag_tracker_context
)


@click.command('time')
@help_option
@click.option(
    "--attach-diff/--no-attach-diff",
    default=True,
    show_default=True,
    required=False,
)
@pass_tag_tracker_context
@pass_create_context
@pass_flow_context
def time_(flow_context, create_context, tag_tracker_context, attach_diff):
    try:
        repository_dir = tag_tracker_context.repository_dir
        git_adapter = GitAdapter(repository_dir)

        version_searcher = TimeBasedVersionSearcher(
            git_adapter
        )

        result = version_searcher.find_current_and_previous_version()
        current_version = result.current_version
        previous_version = result.previous_version
        diff_content = None

        current_commit = current_version.get('commit')
        previous_commit = previous_version.get('commit')

        if previous_commit == current_commit:
            import sys
            click.echo(
                "Previous commit ({0}) and Current commit ({1}) are the same, nothing to do.".format(
                    previous_commit, current_commit
                ),
                file=sys.stderr
            )
            return

        if attach_diff:
            diff_content = git_adapter.diff(
                previous_commit,
                current_commit,
            )

        deploy_metrics = git_adapter.diff_stats(
            previous_commit,
            current_commit,
        ).dict

        flow = flow_context.create_conviso_rest_api_client()

        authors_data = git_adapter.get_commit_authors_by_range(
            previous_commit, current_commit
        )

        deploy = flow.deploys.create(
            current_version=current_version,
            previous_version=previous_version,
            diff_content=diff_content,
            metrics=deploy_metrics,
            project_metrics=project_metrics(repository_dir),
            commit_authors=authors_data
        )

        click.echo(
            create_context.output_formatter.format(deploy)
        )
    except Exception as e:
        on_http_error(e)
        raise click.ClickException(str(e)) from e



import click

from convisoappsec.flowcli import help_option

from .version_tracker import version_tracker


@click.group('with')
@help_option
def with_():
    pass


with_.add_command(version_tracker)

with_.epilog = '''
  Run flow findings create with COMMAND --help for more information
  on a command.
'''



import click
import json
import hashlib
from convisoappsec.flowcli.context import pass_flow_context
from convisoappsec.common.graphql.errors import ResponseError

class ContainerVulnerabilityManager:
    """
    Manages the lifecycle of vulnerabilities detected in container/image scans, including closing and reopening.
    """

    def __init__(self):
        pass

    def close_vulnerability(self):
        """
        Close vulnerabilities on conviso platform when is not detected anymore.
        """
        result_file = "result.json"
        hash_issues = []

        log_func("Running automatic closure of vulnerabilities for image/container...")

        try:
            with open(result_file, 'r') as file:
                data = json.load(file)
                results = data.get('Results', [])

                for result in results:
                    if 'Vulnerabilities' in result and isinstance(result['Vulnerabilities'], list):
                        for vulnerability in result['Vulnerabilities']:
                            title = vulnerability.get('Title', '')
                            description = vulnerability.get('Description', '')
                            severity = vulnerability.get('Severity', '')
                            cve = vulnerability.get('VulnerabilityID', '')
                            package_name = vulnerability.get('PkgName', '')

                            hash_value = generate_hash_issue(title, description, severity, cve, package_name)
                            hash_issues.append(hash_value)
        except FileNotFoundError:
            print(f"Error: The file '{result_file}' was not found.")
        except json.JSONDecodeError:
            print(f"Error: Could not decode JSON from '{result_file}'. Check file format.")
        except Exception as e:
            print(f"An unexpected error occurred: {e}")

        issues_from_conviso = fetch_and_merge_conviso_issues()

        if len(issues_from_conviso) == 0:
            log_func("No vulnerabilities were found on the Conviso Platform!")
            return None

        issues_with_fix_accepted = [item for item in issues_from_conviso if item['status'] == 'FIX_ACCEPTED']
        issues_without_fix_accepted = [item for item in issues_from_conviso if item['status'] != 'FIX_ACCEPTED']
        set_of_hash_issues = set(hash_issues)

        close_issues(issues_from_cp=issues_without_fix_accepted, issues_from_current_scan=set_of_hash_issues)
        reopen_vulnerability(issues_with_fix_accepted=issues_with_fix_accepted, hash_issues=set_of_hash_issues)

        return None

def generate_hash_issue(title, description, severity, cve, package_name):
    """
    Generates a SHA256 hash based on the provided vulnerability details.
    """

    concatenated_string = f"{title}{description}{severity}{cve}{package_name}"

    return hashlib.sha256(concatenated_string.encode('utf-8')).hexdigest()

@pass_flow_context
def close_issues(flow_context, issues_from_cp, issues_from_current_scan):
    """
    method to close container issues on conviso platform

    # issues_from_cp are issues already on conviso platform.
    # issues_from_current_scan are issues identify on each time the ast command runs, these are always from a full code
    base scan
    """
    conviso_api = flow_context.create_conviso_api_client_beta()
    differences = [
        {'id': item['id'], 'originalIssueIdFromTool': item['originalIssueIdFromTool']}
        for item in issues_from_cp if item['originalIssueIdFromTool'] not in issues_from_current_scan
    ]

    if len(differences) == 0:
        log_func("No vulnerabilities have been fixed yet...")
        return

    log_func("Container: Fixing {issues} vulnerabilities on the Conviso Platform...".format(issues=len(differences)))

    for issue in differences:
        issue_id = issue['id']
        status = 'FIX_ACCEPTED'
        reason = ("The vulnerability is no longer found in the specified image. Its status has been updated by "
                  "Conviso AST")

        conviso_api.issues.update_issue_status(issue_id=issue_id, status=status, reason=reason)

@pass_flow_context
def reopen_vulnerability(flow_context, issues_with_fix_accepted, hash_issues):
    """
    Reopen vulnerabilities on conviso platform when is detected and already exists on conviso platform and
    was fixed in other moment.
    """
    conviso_api = flow_context.create_conviso_api_client_beta()

    issues_to_reopen = [
        {'id': item['id'], 'originalIssueIdFromTool': item['originalIssueIdFromTool']}
        for item in issues_with_fix_accepted if item['originalIssueIdFromTool'] in hash_issues
    ]

    if issues_to_reopen:
        log_func("Container: Reopening {issues} vulnerability/vulnerabilities on conviso platform ...".format(
            issues=len(issues_to_reopen))
        )

    for issue in issues_to_reopen:
        issue_id = issue['id']
        status = 'IDENTIFIED'
        reason = 'Status has been updated from Fixed to Identified by Conviso AST'

        conviso_api.issues.update_issue_status(issue_id=issue_id, status=status, reason=reason)

@pass_flow_context
@click.pass_context
def fetch_and_merge_conviso_issues(context, flow_context):
    page = 1
    conviso_api = flow_context.create_conviso_api_client_beta()
    company_id = context.params['company_id']
    asset_id = context.params['asset_id']
    statuses = ['CREATED', 'IDENTIFIED', 'IN_PROGRESS', 'AWAITING_VALIDATION', 'FIX_ACCEPTED']
    merged_issues = []

    while True:
        try:
            issues_from_cp = conviso_api.issues.auto_close_vulnerabilities(
                company_id, asset_id, statuses, page, vulnerability_type='CONTAINER_FINDING'
            )
        except ResponseError as error:
            if 'Variable $company_id' in str(error):
                log_func(f"Invalid company_id passed: {company_id}", fg='red')
            else:
                log_func(f"error: {error}", fg='red')

            log_func(
                "⚠️ Auto-close will not be performed at this time. "
                "Please set it using --company-id and try again.",
                fg='red'
            )
            return None

        total_pages = issues_from_cp['metadata']['totalPages']
        issues_collection = issues_from_cp['collection']

        issues_collection = [
            item for item in issues_collection
            if item.get('scanSource') == 'conviso_scanner'
        ]
        container_issues = [
            item for item in issues_collection
            if item.get('type') == 'CONTAINER_FINDING'
        ]

        if container_issues:
            merged_issues.extend(container_issues)

        if page >= total_pages:
            break
        page += 1

    return merged_issues

def log_func(msg, new_line=True, fg='white'):
    click.echo(click.style(msg, bold=True, fg=fg), nl=new_line, err=True)



import logging

logging.basicConfig(
    filename='output.log',
    filemode='w',
    level=logging.DEBUG,
    format='%(levelname)s:  %(filename)s:  %(threadName)s:  %(name)s:  %(message)s'
)

LOGGER = logging.getLogger(__name__)


def log_and_notify_ast_event(flow_context, company_id, asset_id, ast_log):
    """
    Logs an AST (Application Security Test) event for a specific company and asset,
    and sends notifications.

    This method performs the following actions:
    1. Sends the provided AST log to a monitoring platform like Datadog for logging and tracking.
    2. Sends a Slack notification with relevant details about the AST event.

    Args:
        flow_context: The context of the AST execution.
        company_id (int): The ID of the company for which the AST event is being logged.
        asset_id (int): The ID of the asset associated with the AST event.
        ast_log (string): A string with the log details of the AST event.
    """
    conviso_api = flow_context.create_conviso_graphql_client()
    conviso_api.ast_errors.send_ast_error(company_id = company_id, asset_id = asset_id, log = ast_log)



class CommonException(RuntimeError):
    pass

class ExecutionFailureException(CommonException):
    
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.exit_code = kwargs['exit_code']



from .version_searcher_result import VersionSearcherResult


class TimeBasedVersionSearcher(object):
    def __init__(self, version_control_system_adapter):
        self.version_control_system_adapter = version_control_system_adapter

    def find_current_and_previous_version(self):
        tags = self.version_control_system_adapter.tags()

        current_tag = None
        previous_tag = None

        tags = tags[:2]

        if len(tags) >= 2:
            (current_tag, previous_tag) = tags
        elif len(tags) == 1:
            current_tag = tags[0]
        else:
            raise Exception("Was not possible find the current tag")

        current_commit = self.version_control_system_adapter.show_commit_from_tag(  # noqa: E501
            current_tag
        )

        previous_commit = self.version_control_system_adapter.empty_repository_tree_commit  # noqa: E501

        if previous_tag:
            previous_commit = self.version_control_system_adapter.show_commit_from_tag(  # noqa: E501
                previous_tag
            )

        return VersionSearcherResult(
            current_commit=current_commit,
            previous_commit=previous_commit,
            current_tag=current_tag,
            previous_tag=previous_tag,
        )



from convisoappsec.flow.graphql_api.beta.models.issues.normalize import Normalize


class CreateOrUpdateContainerFindingInput:
    def __init__(
            self,
            asset_id,
            title,
            description,
            severity,
            solution,
            reference,
            affected_version,
            package,
            cve,
            patched_version,
            category,
            original_issue_id_from_tool
    ):
        self.asset_id = asset_id
        self.title = title
        self.description = description
        self.severity = Normalize.normalize_severity(severity)
        self.solution = solution
        self.reference = reference
        self.affected_version = affected_version
        self.package = package
        self.patched_version = patched_version
        self.original_issue_id_from_tool = original_issue_id_from_tool
        self.category = self.process_field(category)
        self.cve = self.process_field(cve)

    def to_graphql_dict(self):
        """
        This function returns a dictionary containing various attributes of an
        asset in a GraphQL format.
        """
        return {
            "assetId": int(self.asset_id),
            "title": self.title,
            "description": self.description,
            "severity": self.severity,
            "solution": self.solution,
            "reference": self.reference,
            "affectedVersion": self.affected_version,
            "package": self.package,
            "cve": self.cve,
            "patchedVersion": self.patched_version,
            "category": self.category,
            "originalIssueIdFromTool": self.original_issue_id_from_tool
        }

    @staticmethod
    def process_field(value):
        """
        Processes a field to ensure it is converted into a string.

        - If the value is a list, it joins the items into a comma-separated string.
        - If the value is a string, it returns the string as is.
        - If the value is neither a list nor a string, it returns an empty string.

        Args:
            value (list | str | Any): The value to process.

        Returns:
            str: The processed string representation of the value.
        """
        if isinstance(value, list):
            return ' , '.join(value)
        elif isinstance(value, str):
            return value
        return ''



import sys
import click
import traceback
import time
import json
from copy import deepcopy as clone
from base64 import b64decode
from re import search as regex_search
from convisoappsec.flow import GitAdapter
from convisoappsec.flowcli import help_option
from convisoappsec.flowcli.common import (asset_id_option, on_http_error)
from convisoappsec.flowcli.context import pass_flow_context
from convisoappsec.sast.decision import Decision, Severity
from convisoappsec.sast.sastbox import SASTBox
from convisoappsec.flowcli.requirements_verifier import RequirementsVerifier
from docker.errors import APIError
from convisoappsec.logger import LOGGER
from convisoappsec.common.cleaner import Cleaner


def log_func(msg, new_line=True):
    click.echo(msg, nl=new_line, err=True)


def find_blocked_issues(
        results_filepaths, overall_threshold, severity_threshold, severity
):
    blocked_issues = []

    for result_path in results_filepaths:
        report_decision = Decision(result_path)

        overall_policy = report_decision.block_from_findings(overall_threshold)
        severity_policy = report_decision.block_from_severity(
            severity, severity_threshold
        )

        if overall_policy or severity_policy:
            blocked_issues.append(report_decision)
            click.echo(
                "Failing execution due one or more blocking flags", file=sys.stderr
            )

        click.echo(result_path)

    return blocked_issues


def print_blocked_issues(blocked_issues):
    for issue_decision in blocked_issues:
        issues = issue_decision.filtered_issues()

        for issue in issues:
            click.echo(
                "{issue_name}\n{filepath} at line {line_index}\n".format(
                    issue_name=issue["title"],
                    filepath=issue["filename"],
                    line_index=issue["line"],
                )
            )
    pass


def perform_sastbox_scan(
    conviso_rest_api, sastbox_registry, sastbox_repository_name, sastbox_tag, sastbox_skip_login, repository_dir, end_commit, start_commit, logger
):

    max_retries = 5
    retries = 0
    sastbox = SASTBox(registry=sastbox_registry, repository_name=sastbox_repository_name, tag=sastbox_tag)
    pull_progress_bar = click.progressbar(length=sastbox.size, label="Performing SAST download...")

    while retries < max_retries:
        try:
            if not sastbox_skip_login:
                logger("Checking SASTBox authorization...")
                token = conviso_rest_api.docker_registry.get_sast_token()
                sastbox.login(token)

            with pull_progress_bar as progressbar:
                for downloaded_chunk in sastbox.pull():
                    progressbar.update(downloaded_chunk)
            break
        except APIError as e:
            retries += 1
            logger(f"Retrying {retries}/{max_retries}...")
            time.sleep(1)

            if retries == max_retries:
                logger("Max retries reached. Failed to perform SAST download.")
                raise Exception(f"Max retries reached. Could not complete the SAST download. Error: {str(e)}")

    logger("Starting SAST scan diff...")

    reports = sastbox.run_scan_diff(repository_dir, end_commit, start_commit, log=logger)

    logger("SAST scan diff done.")

    results_filepaths = []
    for r in reports:
        try:
            file_path = str(r)
            results_filepaths.append(file_path)
        except Exception as e:
            click.echo(f"Error decoding file path: {r} with error {e}.", file=sys.stderr)

    return results_filepaths


def parse_conviso_references(references=[]):
    divider = "\n"

    references_to_join = []

    for reference in references:
        if reference:
            references_to_join.append(reference)

    return divider.join(references_to_join)


def parse_code_snippet(encoded_base64):
    try:
        decoded_text = b64decode(encoded_base64).decode("utf-8")
    except UnicodeDecodeError:
        try:
            decoded_text = b64decode(encoded_base64, validate=False).decode("latin-1")
        except Exception as e:
            print("Error handling decoding error:", e, file=sys.stderr)
            decoded_text = ''

    lines = decoded_text.split("\n")

    cleaned_lines = []
    for line in lines:
        cleaned_line = line.split(": ", 1)[-1]
        cleaned_lines.append(cleaned_line)

    code_snippet = "\n".join(cleaned_lines)

    return code_snippet


def parse_first_line_number(encoded_base64):
    try:
        decoded_text = b64decode(encoded_base64).decode("utf-8")
    except UnicodeDecodeError:
        try:
            decoded_text = b64decode(encoded_base64, validate=False).decode("latin-1")
        except Exception as e:
            print("Error handling decoding error:", e, file=sys.stderr)
            decoded_text = ''

    regex = r"^(\d+):"

    result = regex_search(regex, decoded_text)

    if result and result.group(1):
        return result.group(1)

    line_number_when_not_found = 1
    return line_number_when_not_found


def deploy_results_to_conviso_beta(
        flow_context, results_filepaths, asset_id, company_id, commit_ref=None, deploy_id=None, control_sync_status_id=None
):
    """Send SAST results to the Conviso platform."""

    results_context = click.progressbar(
        results_filepaths, label="🔧 Sending SAST report to the Conviso Platform ..."
    )

    with results_context as reports:
        for report_path in reports:
            api_key = flow_context.key
            max_retries = 5
            conviso_api = flow_context.create_conviso_graphql_client()

            for attempt in range(1, max_retries + 1):
                try:
                    conviso_api.issues.send_issues_file(
                        company_id=company_id, asset_id=asset_id, file_path=report_path, api_key=api_key,
                        vulnerability_type='SAST_FINDING',
                        deploy_id=deploy_id, commit_ref=commit_ref, control_sync_status_id=control_sync_status_id
                    )
                    break
                except Exception as e:
                    if attempt == max_retries:
                        print(f"Failed after {max_retries} attempts: {e}")
                    else:
                        wait_time = 2 ** attempt  # Exponential backoff (2s, 4s, 8s, 16s)
                        print(f"Retry {attempt}/{max_retries} failed. Retrying in {wait_time} seconds...")
                        time.sleep(wait_time)


@click.command()
@asset_id_option(required=False)
@click.option(
    "-s",
    "--start-commit",
    required=False,
    help="If no value is set so the empty tree hash commit is used.",
)
@click.option(
    "-e",
    "--end-commit",
    required=False,
    help="""If no value is set so the HEAD commit
    from the current branch is used""",
)
@click.option(
    "-r",
    "--repository-dir",
    default=".",
    show_default=True,
    type=click.Path(
        exists=True,
        resolve_path=True,
    ),
    required=False,
    help="The source code repository directory.",
)
@click.option(
    "--fail-on-severity-threshold",
    required=False,
    help="If the threshold of the informed severity and higher has reach, then the command will fail after send the results to AppSec Flow.\n \
    The severity levels are: UNDEFINED, INFO, LOW, MEDIUM, HIGH, CRITICAL.",
    type=click.Tuple([str, int]),
    default=(None, None),
)
@click.option(
    "--fail-on-threshold",
    required=False,
    help="If the threshold has reach then the command will fail after send the result to AppSec Flow",
    type=int,
    default=False,
)
@click.option(
    "--send-to-flow/--no-send-to-flow",
    default=True,
    show_default=True,
    required=False,
    hidden=True,
    help="""Enable or disable the ability of send analysis result
    reports to flow.""",
)
@click.option(
    "--deploy-id",
    default=None,
    required=False,
    hidden=True,
    envvar=("CONVISO_DEPLOY_ID", "FLOW_DEPLOY_ID")
)
@click.option(
    "--sastbox-registry",
    default="",
    required=False,
    hidden=True,
    envvar=("CONVISO_SASTBOX_REGISTRY", "FLOW_SASTBOX_REGISTRY"),
)
@click.option(
    "--sastbox-repository-name",
    default="",
    required=False,
    hidden=True,
    envvar=("CONVISO_SASTBOX_REPOSITORY_NAME", "FLOW_SASTBOX_REPOSITORY_NAME"),
)
@click.option(
    "--sastbox-tag",
    default=SASTBox.DEFAULT_TAG,
    required=False,
    hidden=True,
    envvar=("CONVISO_SASTBOX_TAG", "FLOW_SASTBOX_TAG"),
)
@click.option(
    "--sastbox-skip-login/--sastbox-no-skip-login",
    default=False,
    required=False,
    hidden=True,
    envvar=("CONVISO_SASTBOX_SKIP_LOGIN", "FLOW_SASTBOX_SKIP_LOGIN"),
)
@click.option(
    '--experimental',
    default=False,
    is_flag=True,
    hidden=True,
    help="Enable experimental features.",
)
@click.option(
    "--company-id",
    required=False,
    envvar=("CONVISO_COMPANY_ID", "FLOW_COMPANY_ID"),
    help="Company ID on Conviso Platform",
)
@click.option(
    '--asset-name',
    required=False,
    envvar=("CONVISO_ASSET_NAME", "FLOW_ASSET_NAME"),
    help="Provides a asset name.",
)
@click.option(
    '--vulnerability-auto-close',
    default=False,
    is_flag=True,
    hidden=True,
    help="Enable auto fixing vulnerabilities on cp.",
)
@click.option(
    '--from-ast',
    default=False,
    is_flag=True,
    hidden=True,
    help="Internal use only.",
)
@click.option(
    '--cleanup',
    default=False,
    is_flag=True,
    show_default=True,
    help="Clean up system resources, including temporary files, stopped containers, unused Docker images and volumes.",
)
@click.option(
    '--control-sync-status-id',
    required=False,
    hidden=True,
    help="Control sync status id.",
)
@help_option
@pass_flow_context
@click.pass_context
def run(context, flow_context, asset_id, company_id, end_commit, start_commit, repository_dir, send_to_flow, deploy_id,
        sastbox_registry, sastbox_repository_name, sastbox_tag, sastbox_skip_login, fail_on_threshold, fail_on_severity_threshold,
        experimental, asset_name, vulnerability_auto_close, from_ast, cleanup, control_sync_status_id):
    """
    This command will perform SAST analysis at the source code. The analysis
    results can be reported or not to flow application. The analysis can be
    applied to specific commit range.

    This command will write the analysis reports files paths to stdout.
    """
    if not from_ast:
        prepared_context = RequirementsVerifier.prepare_context(clone(context))

        params_to_copy = [
            'asset_id', 'company_id', 'start_commit', 'end_commit',
            'repository_dir', 'send_to_flow', 'deploy_id', 'sastbox_registry',
            'sastbox_repository_name', 'sastbox_tag', 'sastbox_skip_login',
            'experimental', 'asset_name', 'vulnerability_auto_close', 'company_id',
            'cleanup'
        ]

        for param_name in params_to_copy:
            context.params[param_name] = (
                    locals()[param_name] or prepared_context.params[param_name]
            )

    perform_command(context, flow_context, context.params['asset_id'], context.params['deploy_id'], context.params['end_commit'],
                    context.params['start_commit'], context.params['repository_dir'], context.params['send_to_flow'], context.params['sastbox_registry'],
                    context.params['sastbox_repository_name'], context.params['sastbox_tag'], context.params['sastbox_skip_login'],
                    context.params['fail_on_threshold'], context.params['fail_on_severity_threshold'], context.params['experimental'],
                    context.params['company_id'], context.params['cleanup'], from_ast, control_sync_status_id)

def perform_command(context, flow_context, asset_id, deploy_id, end_commit, start_commit, repository_dir, send_to_flow, sastbox_registry, sastbox_repository_name, sastbox_tag,
                    sastbox_skip_login, fail_on_threshold, fail_on_severity_threshold, experimental, company_id, cleanup, from_ast, control_sync_status_id):

    if send_to_flow and experimental and not asset_id:
        raise click.MissingParameter(
            "It is required when sending reports to Conviso Platform using experimental API.",
            param_type="option",
            param_hint="--asset-id",
        )

    severity, severity_threshold = fail_on_severity_threshold
    overall_threshold = fail_on_threshold

    if severity and not Severity.has_value(severity):
        raise click.BadParameter(
            "{} is not a valid Severity. Use a valid Severity value:\n{}".format(
                severity, [severity.name for severity in Severity]
            ),
            param_hint="--fail-on-severity-threshold",
        )

    try:
        git_adapter = GitAdapter(repository_dir)

        end_commit = end_commit or git_adapter.head_commit

        start_commit = start_commit or git_adapter.empty_repository_tree_commit

        if start_commit == end_commit:
            click.echo(
                "Previous commit ({0}) and current commit ({1}) are the same; nothing to do.".format(
                    start_commit, end_commit
                ),
                file=sys.stderr,
            )
            return

        conviso_rest_api = flow_context.create_conviso_rest_api_client()

        results_filepaths = perform_sastbox_scan(
            conviso_rest_api, sastbox_registry, sastbox_repository_name, sastbox_tag, sastbox_skip_login, repository_dir, end_commit, start_commit, log_func,
        )

        # this set total sast vulnerability found in sast scan to send to control sync status
        context.params['sast_vulnerability_count'] = total_vulnerability_count(results_filepaths[0])

        if send_to_flow:
            deploy_results_to_conviso_beta(
                flow_context, results_filepaths, asset_id, company_id, commit_ref=end_commit, deploy_id=deploy_id, control_sync_status_id=control_sync_status_id
            )

        blocked_issues = find_blocked_issues(
            results_filepaths, overall_threshold, severity_threshold, severity
        )

        if blocked_issues:
            print_blocked_issues(blocked_issues)
            sys.exit(1)

        if cleanup and from_ast == False:
            LOGGER.info("🧹 Cleaning up ...")
            cleaner = Cleaner()
            cleaner.cleanup()

    except Exception as e:
        traceback.print_exc()
        on_http_error(e)
        raise click.ClickException(str(e)) from e

def total_vulnerability_count(file_path: str) -> int:
    """
    Extract the total vulnerability count from a SAST scan result file.

    Args:
        file_path (str): Path to JSON result file containing vulnerability scan results.
                        The file should have a 'summary' section with
                        'issues_count.total' field.

    Returns:
        int: Total number of vulnerabilities found in the scan result file.
             Returns 0 if the file doesn't exist or no vulnerabilities are found.
    """

    try:
        with open(file_path, 'r', encoding='utf-8') as f:
            data = json.load(f)
            return data['summary']['issues_count']['total']
    except (FileNotFoundError, KeyError, json.JSONDecodeError):
        return 0

EPILOG = """
Examples:

  \b
  1 - Reporting the results to flow api:
    1.1 - Running an analysis at all commit range:
      $ export CONVISO_API_KEY='your-api-key'
      $ {command}

    \b
    1.2 - Running an analysis at specific commit range:
      $ export CONVISO_API_KEY='your-api-key'
      $ {command} --start-commit "$(git rev-parse HEAD~5)" --end-commit "$(git rev-parse HEAD)"

    \b
  2 - Using flags to break the job on findings ocurrence:
    2.1 - Running an analysis and break the build if there is 10 findings or more:
      $ export CONVISO_API_KEY='your-api-key'
      $ {command} --fail-on-threshold 10
    \b
    2.2 - Running an analysis and break the build if there is 5 findings with HIGH severity or higher
      $ export CONVISO_API_KEY='your-api-key'
      $ {command} --fail-on-severity-threshold HIGH 5
"""  # noqa: E501

SHORT_HELP = "Perform SAST analysis"

command = "conviso sast run"
run.short_help = SHORT_HELP
run.epilog = EPILOG.format(
    command=command,
)
